import typing, base64
from xml.etree.ElementTree import Element as XML_Element

from .exception import CleanintyExceptionBase

__all__ = [
	"XmlParseHelper",
	"XMLParseError",
	"XML_Element"
]

class XMLParseError(CleanintyExceptionBase):
	"""General XML parse error"""

class XmlParseHelper:
	_xml_response_namespaces: typing.Dict[str, str] = {}

	@staticmethod
	def _xml_raise_if_any_subelement(element: XML_Element) -> None:
		if element.find('./{*}*') is not None:
			raise XMLParseError("Subelements found where they are not wanted")

	@staticmethod
	def _xml_raise_if_text(element: XML_Element) -> None:
		if element.text is not None:
			raise XMLParseError("Element with unexpected text")

	@staticmethod
	def _xml_get_element(
		parent: 'XmlParseHelper',
		element: XML_Element
	) -> XML_Element:
		return element

	@staticmethod
	def _xml_get_str_element(
		parent: 'XmlParseHelper',
		element: XML_Element
	) -> str:
		parent._xml_raise_if_any_subelement(element)
		if element.text is None:
			return ''
		return element.text

	@staticmethod
	def _xml_get_base64_element(
		parent: 'XmlParseHelper',
		element: XML_Element
	) -> bytes:
		text = parent._xml_get_str_element(parent, element)
		return base64.b64decode(text + '===')

	@staticmethod
	def _xml_get_int_element(
		parent: 'XmlParseHelper',
		element: XML_Element
	) -> int:
		parent._xml_raise_if_any_subelement(element)
		if element.text is None:
			return 0
		return int(element.text)

	@staticmethod
	def _xml_get_int_base16_element(
		parent: 'XmlParseHelper',
		element: XML_Element
	) -> int:
		parent._xml_raise_if_any_subelement(element)
		if element.text is None:
			return 0
		return int(element.text, 16)

	def _xml_element_parse(
		self,
		element: XML_Element,
		tagname: str,
		parser: typing.Callable[
			[
				'XmlParseHelper',
				XML_Element
			],
			typing.Any
		],
		optional: bool = False
	) -> typing.Any:
		element = element.find('./{0}'.format(tagname), self._xml_response_namespaces)

		if element is None and not optional:
			raise XMLParseError("Non-optional element not found: {0}".format(tagname))
		elif element is None:
			return None

		return parser(self, element)

	def _xml_multi_element_parse(
		self,
		element: XML_Element,
		tagname: str,
		parser: typing.Callable[
			[
				'XmlParseHelper',
				XML_Element
			],
			typing.Any
		],
		optional: bool = False
	) -> typing.Any:
		_iter = element.iterfind('./{0}'.format(tagname), self._xml_response_namespaces)

		i = None
		values = []
		for i, j in enumerate(_iter):
			values.append(parser(self, j))

		if i is None and not optional:
			raise XMLParseError("Non-optional element not found: {0}".format(tagname))
		elif i is None:
			return None

		return values

	def _xml_multi_element_parse_ret_none(
		self,
		element: XML_Element,
		tagname: str,
		parser: typing.Callable[
			[
				'XmlParseHelper',
				XML_Element
			],
			None
		],
		optional: bool = False
	) -> bool:
		_iter = element.iterfind('./{0}'.format(tagname), self._xml_response_namespaces)

		i = None
		for i, j in enumerate(_iter):
			parser(self, j)

		if i is None and not optional:
			raise XMLParseError("Non-optional element not found: {0}".format(tagname))

		return i is not None # let us know if we got something at least



import time, random
from enum import IntEnum, unique

from . import ias, ecs
from .manager import CtrSoapManager, ServiceLevel
from .exception import CTRExceptionBase

__all__ = [
	"CtrSoapCheckRegister",
	"CtrSoapSessionInit",
	"CtrSoapSessionConnect",
	"SysApps",
	"CtrSoapUseSystemApps",
	"HelperError"
]

class HelperError(CTRExceptionBase):
	"""General Helper error exception"""

def CtrSoapCheckRegister(soap_device: CtrSoapManager) -> None:
	if not isinstance(soap_device, CtrSoapManager):
		raise HelperError("Expected CtrSoapManager")

	status = ecs.GetAccountStatus(soap_device)

	status.update_soap_managers_uris()

	if soap_device.service_standby and not soap_device.is_service_ready():
		raise HelperError("Unhandled combination of service status")

	if status.errorcode == 902 or status.errorcode == 903:
		reginfo = ias.GetRegistrationInfo(soap_device, ias.GetChallenge(soap_device).challenge)

		if not reginfo.is_registered:
			soap_device.unregister_account()

		else:
			soap_device.register_account('R', reginfo.devicetoken, reginfo.accountid, reginfo.country, reginfo.extaccountid, reginfo.devicecode, reginfo.devicetokenexpired)

	elif status.errorcode != 901:
		status.validate_errorcode()

		if status.accountstatus == 'U':
			soap_device.unregister_account()
		else:
			# we are in abnormal condition compared to the actual console
			# we are foreign
			# it may change its condition
			soap_device.register_account(status.accountstatus, soap_device.st_token, status.accountid, status.country, soap_device.ext_account_id, soap_device.device_code, soap_device.register_expired)


def CtrSoapSessionInit(soap_device: CtrSoapManager) -> None:
	"""
	Partial recommended session start for EC, complete for NUP.
	It's recommended to set an AppId and TIN to the manager.
	"""

	if not isinstance(soap_device, CtrSoapManager):
		raise HelperError("Expected CtrSoapManager")

	reginfo = None

	status = ecs.GetAccountStatus(soap_device)

	status.update_soap_managers_uris()

	if soap_device.service_standby and not soap_device.is_service_ready():
		raise HelperError("Unhandled combination of service status")

	if status.errorcode == 902 or status.errorcode == 903:
		reginfo = ias.GetRegistrationInfo(soap_device, ias.GetChallenge(soap_device).challenge)

		if not reginfo.is_registered:
			soap_device.unregister_account()

		elif reginfo.devicetokenexpired:
			soap_device.country_change(reginfo.country)

			reginfo = ias.Register(soap_device, ias.GetChallenge(soap_device).challenge)

			soap_device.register_account('R', reginfo.devicetoken, reginfo.accountid, reginfo.country, reginfo.extaccountid, reginfo.devicecode)

		else:
			soap_device.register_account('R', reginfo.devicetoken, reginfo.accountid, reginfo.country, reginfo.extaccountid, reginfo.devicecode)

		if soap_device.service_level == ServiceLevel.SHOP:
			status = ecs.GetAccountStatus(soap_device)

			status.update_soap_managers_uris()

	elif status.errorcode != 901:
		status.validate_errorcode()

	if status.accountstatus == 'P':
		if reginfo is not None:
			soap_device.country_change(reginfo.country)

		reginfo = ias.AccountTransfer(soap_device, ias.GetChallenge(soap_device).challenge)

		soap_device.register_account('T', reginfo.devicetoken, reginfo.accountid, reginfo.country, reginfo.extaccountid, reginfo.devicecode)

	elif status.accountstatus == 'R' or status.accountstatus == 'T':
		soap_device.register_account(status.accountstatus, soap_device.st_token, soap_device.account_id, soap_device.country, soap_device.ext_account_id, soap_device.device_code)
	elif status.accountstatus == 'U':
		reginfo = ias.Register(soap_device, ias.GetChallenge(soap_device).challenge)

		soap_device.register_account('R', reginfo.devicetoken, reginfo.accountid, reginfo.country, reginfo.extaccountid, reginfo.devicecode)
	else:
		raise HelperError("Unknown account status indicator returned from SOAP!")

	if soap_device.service_level == ServiceLevel.SHOP:
		status = ecs.GetAccountStatus(soap_device)

		status.update_soap_managers_uris()

		soap_device.set_ivs_sync_flag(status.ivssyncflag)

	if soap_device.service_standby and not soap_device.is_service_ready():
		raise HelperError("Unhandled combination of service status")

# post transfer connect, import_remote_ivs = True
def CtrSoapSessionConnect(soap_device: CtrSoapManager, import_remote_ivs: bool = False, drop_remote_ivs: bool = True, *, forced_ivs_remote_import: bool = False, forced_ivs_send: bool = False):
	"""
	More complete init for EC manager.
	It's recommended to set an AppId and TIN to the manager.
	"""

	if forced_ivs_remote_import:
		import_remote_ivs = True

	if not isinstance(soap_device, CtrSoapManager):
		raise HelperError("Expected CtrSoapManager")

	CtrSoapSessionInit(soap_device)

	if soap_device.service_level != ServiceLevel.SHOP:
		return

	if soap_device.service_standby:
		return

	try:
		current_ivs = soap_device.export_ivs()
	except Exception:
		return

	if (soap_device.ivs_sync_flag or forced_ivs_remote_import) and not forced_ivs_send:
		if import_remote_ivs:
			remote_ivs = ias.GetIVSData(soap_device).ivsdata

			if not remote_ivs and soap_device.ivs_sync_flag:
				ias.ReportIVSSync(soap_device)

			else:
				imported = False

				try:
					soap_device.import_ivs(remote_ivs)
					imported = True
				except Exception:
					pass

				if imported and soap_device.ivs_sync_flag:
					try:
						ias.ReportIVSSync(soap_device)
					except Exception:
						soap_device.import_ivs(current_ivs)
						raise

				if imported:
					soap_device.set_last_sync_ivs(remote_ivs)

		elif drop_remote_ivs:
			ias.ReportIVSSync(soap_device)

	elif soap_device.is_msed_signed and \
		(soap_device.account_status == 'R' or soap_device.account_status == 'T'):
		last_ivs = soap_device.last_sync_ivs

		if forced_ivs_send or not last_ivs or current_ivs != last_ivs:
			ias.SetIVSData(soap_device, current_ivs)

			soap_device.set_last_sync_ivs(current_ivs)

@unique
class SysApps(IntEnum):
	NIM = 0
	ESHOP = 1
	SYSTRANSFER = 2

def CtrSoapUseSystemApps(soap_device: CtrSoapManager, app: SysApps) -> None:
	if not isinstance(soap_device, CtrSoapManager):
		raise HelperError("Expected CtrSoapManager")

	if app == SysApps.NIM:
		soap_device.set_app_and_tin(0x0004013000002C02, 1234)

	elif app == SysApps.ESHOP:
		tid = {
			'JPN': 0x0004001000020900,
			'USA': 0x0004001000021900,
			'EUR': 0x0004001000022900,
			'KOR': 0x0004001000027900,
			'TWN': 0x0004001000028900
		}.get(soap_device.region, None)

		if tid is None:
			raise HelperError("eShop not available on this region or region invalid")

		soap_device.set_app_and_tin(tid, 56789)

	elif app == SysApps.SYSTRANSFER:
		tid = {
			'JPN': 0x0004001000020A00,
			'USA': 0x0004001000021A00,
			'EUR': 0x0004001000022A00,
			'KOR': 0x0004001000027A00,
			'TWN': 0x0004001000028A00
		}.get(soap_device.region, None)

		if tid is None:
			raise HelperError("System transfer not available on this region or region invalid")

		soap_device.set_app_and_tin(tid, 1111)



from struct import unpack
import typing, hashlib, json, re, threading, os

from cryptography.hazmat.primitives.asymmetric import rsa, ec
from cryptography.hazmat.primitives import serialization
from cryptography import x509
from cryptography.hazmat.backends import default_backend
import cryptography.hazmat.primitives.ciphers as crypto_ciphers

from .environment import find_in_config_dirs

# in the future, I want this to load any possible constant from the system's extracted data
# mainly for sake of building a ctr_constant.json that can be used by any application that wishes for system data

# ACT

_constants_json_lock = threading.RLock()

_act_cli_data = {
	'CLIENT_ID': None,
	'CLIENT_SECRET': None
}

_act_cli_hashes = {
	'25c20a9cbd5612a7f31ea45d27c18fe27bce121353daf90aaf0407cf08356769': 'CLIENT_ID',
	'c8c4ad2420d2ba824eedda113de77a8b662406b972952b64889d76cf71088c01': 'CLIENT_SECRET'
}

def _load_act_secrets(codebin: typing.Optional[bytes] = None) -> None:
	global _act_cli_data

	strs = []

	paths = find_in_config_dirs('ctr_constants.json')

	for i in paths:
		try:
			with open(i, 'r', encoding='utf-8') as j, _constants_json_lock:
				data = json.load(j)
			act = data.get('ACT', {})
			for j in ['CLIENT_ID', 'CLIENT_SECRET']:
				n = act.get(j, None)
				if n is not None and len(n) == 32:
					strs.append(n.encode('ascii'))
		except Exception:
			continue

	if codebin:
		strs += re.findall(b'[0-9a-f]{32}', codebin)

	for i in strs:
		_hash = hashlib.sha256(i).hexdigest()
		name = _act_cli_hashes.get(_hash, None)
		if name is None:
			continue
		_act_cli_data[name] = i.decode('ascii')

	if codebin:
		_save_act_secrets()

def _save_act_secrets() -> None:
	paths = find_in_config_dirs('ctr_constants.json', True)

	for i in paths:
		try:
			with open(i, 'r+', encoding='utf-8') as j, _constants_json_lock:
				try:
					data = json.load(j)
				except json.decoder.JSONDecodeError:
					data = {}
				j.seek(0)
				act = data.get('ACT', {})
				ci = _act_cli_data.get('CLIENT_ID', None)
				cs = _act_cli_data.get('CLIENT_SECRET', None)
				act['CLIENT_ID'] = ci if ci is not None else act.get('CLIENT_ID', None)
				act['CLIENT_SECRET'] = cs if cs is not None else act.get('CLIENT_SECRET', None)
				data['ACT'] = act
				out = json.dumps(data, indent=2)
				j.truncate()
				j.write(out)

		except Exception:
			continue

# CFG

_cfg_pubnums = {
	'LFCS_DEV': None,
	'LFCS_RETAIL': None,
	'SECINFO_DEV': None,
	'SECINFO_RETAIL': None
}

_cfg_key_hashes = {
	'b684323ea8b7d5966048f0c858fa066b3a5219bdef3eb84652b0951722eede4e': 'LFCS_DEV',
	'6dbf136edb9fe5b68e5681ad84ce20634db3cbacf53a025d62ad274dbac5fcfe': 'LFCS_RETAIL',
	'372889b6bee8c7e83cc4b7f5df245ccd04ff0a36fce8cebf1a85a26192b20022': 'SECINFO_DEV',
	'e109d0cce7298a46255bea27240f1630ac6cdaccb1b47a90d613171ac2ddfdda': 'SECINFO_RETAIL'
}

def _load_cfg_rsa_keys(codebin: typing.Optional[bytes] = None) -> None:
	global _cfg_pubnums

	off = -1
	keys = []

	e = 0x10001

	while codebin:
		_off = codebin[off+1:].find(b'\x30\x82\x01\x22')
		if _off == -1:
			break
		off = off + 1 + _off
		try:
			key = serialization.load_der_public_key(codebin[off:])
			if isinstance(key, rsa.RSAPublicKey) or isinstance(key, rsa.RSAPublicKeyWithSerialization):
				keys.append(key.public_numbers())
		except Exception:
			pass

	paths = find_in_config_dirs('ctr_constants.json')

	for i in paths:
		try:
			with open(i, 'r', encoding='utf-8') as j, _constants_json_lock:
				data = json.load(j)
			cfg = data.get('CFG', {})
			for j in ['LFCS_DEV_N', 'LFCS_RETAIL_N', 'SECINFO_DEV_N', 'SECINFO_RETAIL_N']:
				n = cfg.get(j, None)
				try:
					n = int(n)
				except Exception:
					continue
				keys.append(rsa.RSAPublicNumbers(e, n))
		except Exception:
			continue

	for i in keys:
		_hash = hashlib.sha256(hex(i.n).encode('ascii')).hexdigest()
		name = _cfg_key_hashes.get(_hash, None)
		if name is None:
			continue
		_cfg_pubnums[name] = rsa.RSAPublicNumbers(e, i.n)

	if codebin: # save only if potential new keys
		_save_cfg_rsa_keys()

def _save_cfg_rsa_keys() -> None:
	paths = find_in_config_dirs('ctr_constants.json', True)

	for i in paths:
		try:
			with open(i, 'r+', encoding='utf-8') as j, _constants_json_lock:
				try:
					data = json.load(j)
				except json.decoder.JSONDecodeError:
					data = {}
				j.seek(0)
				cfg = data.get('CFG', {})
				ld = _cfg_pubnums.get('LFCS_DEV', None)
				lr = _cfg_pubnums.get('LFCS_RETAIL', None)
				sd = _cfg_pubnums.get('SECINFO_DEV', None)
				sr = _cfg_pubnums.get('SECINFO_RETAIL', None)
				cfg['LFCS_DEV_N'] = f"{ld.n}" if ld is not None else cfg.get('LFCS_DEV_N', None)
				cfg['LFCS_RETAIL_N'] = f"{lr.n}" if lr is not None else cfg.get('LFCS_RETAIL_N', None)
				cfg['SECINFO_DEV_N'] = f"{sd.n}" if sd is not None else cfg.get('SECINFO_DEV_N', None)
				cfg['SECINFO_RETAIL_N'] = f"{sr.n}" if sr is not None else cfg.get('SECINFO_RETAIL_N', None)
				data['CFG'] = cfg
				out = json.dumps(data, indent=2)
				j.truncate()
				j.write(out)

		except Exception:
			continue

# P9

_p9_pubnums = {
	'CDN_TIK_WRAP': None,
	'ROOT_CERT_DEV': None,
	'ROOT_CERT_RETAIL': None,
	'CTR2_DEV': None,
	'CTR2_PROD': None
}

_p9_keys = {
	'IVS_CMAC_KEYY': None,
	'IVS_KEYY': None
}

_p9_key_hashes = {
	'a871c8f16cfc55508d57bfa5741dd8d44d5bc46c9a5361ea12b11007447ea1ae': 'CDN_TIK_WRAP',
	'6dbf136edb9fe5b68e5681ad84ce20634db3cbacf53a025d62ad274dbac5fcfe': 'LFCS_RETAIL',
	'25de2e150fa2f50c5b5e1ba3ce59e3759bbc97a401d9f22cd7a0189b52ffb9a7': 'ROOT_CERT_DEV',
	'73f65a898c67bd8e166899536cfc8f479792ecb9a279c2b8ecd214032511080a': 'ROOT_CERT_RETAIL',
	'39d33bd0c05949d1c3f98a088ac518dac73496411b5a054ff1698b682469adbd': 'CTR2_DEV_X',
	'1fdbeb696cc779f1a04f4ae99fb335ed3e36dadd8e4163d6ae8fbca94205b9cd': 'CTR2_DEV_Y',
	'77752904569351b4d895679b1dd24dd3150a7b1bed2be063ef9e5ff578d3c862': 'CTR2_PROD_X',
	'690dcb701cc12a7e0e67ad57714a608c4eb3b893bf2a648defded0c3a2f702fd': 'CTR2_PROD_Y',
	'c318d16d02c3bd7d21b9239aa7632eab412ffc08353c293f4d28345843052b3c': 'IVS_CMAC_KEYY',
	'fc8e6cc6afc33983a9d377a93031e3d1d2e37810392667c671b76ebc2abe2573': 'IVS_KEYY'
}

def _load_p9_keys(codebin: typing.Optional[bytes] = None) -> None:
	global _p9_pubnums
	global _p9_keys

	off = -1
	keys = []
	certs = []
	aes_keys = []

	e = 0x10001

	while codebin:
		_off = codebin[off+1:].find(b'\x30\x82\x01\x22')
		if _off == -1:
			break
		off = off + 1 + _off
		try:
			key = serialization.load_der_public_key(codebin[off:])
			if isinstance(key, rsa.RSAPublicKey) or isinstance(key, rsa.RSAPublicKeyWithSerialization):
				keys.append(key.public_numbers())
		except Exception:
			pass

	off = -1

	while codebin:
		_off = codebin[off+1:].find(b'\x30\x82\x03')
		if _off == -1:
			break
		off = off + 1 + _off
		try:
			size = unpack(">2xH", codebin[off:off+4])[0] + 4
			cert = x509.load_der_x509_certificate(codebin[off:off+size])
			if not isinstance(cert, x509.Certificate):
				continue
			certs.append(cert)
			# uncertain if different versions demonstrate diff alignments for keys right after one of these certs
			size_a4  = (size + 3) & 0xFFFC
			size_a8  = (size + 7) & 0xFFF8
			size_a16 = (size + 15) & 0xFFF0
			size_a32 = (size + 31) & 0xFFE0
			size_a64 = (size + 63) & 0xFFC0
			sizes = list(set([size_a4, size_a8, size_a16, size_a32, size_a64]))
			for i in sizes:
				aes_keys += [int.from_bytes(j, 'big') for j in unpack("16s16s", codebin[off+i:off+i+32])]
		except Exception:
			pass

	off = -1

	while codebin:
		_off = codebin[off+1:].find(b'\x01\x00\x01\x00\x01\x00\x01\x00')
		if _off == -1:
			break
		off = off + 1 + _off
		try:
			k1 = int.from_bytes(codebin[off+8:off+8+512], 'big')
			k2 = int.from_bytes(codebin[off+8+512:off+8+1024], 'big')
			keys.append(rsa.RSAPublicNumbers(e, k1))
			keys.append(rsa.RSAPublicNumbers(e, k2))
		except Exception:
			pass

	cert_keys = {}

	paths = find_in_config_dirs('ctr_constants.json')

	for i in paths:
		try:
			with open(i, 'r', encoding='utf-8') as j, _constants_json_lock:
				data = json.load(j)
			p9 = data.get('P9', {})
			p9_asym = p9.get('Asymmetric', {})
			p9_aes = p9.get('AES', {})

			for j in ['CDN_TIK_WRAP_N', 'ROOT_CERT_DEV_N', 'ROOT_CERT_RETAIL_N']:
				n = p9_asym.get(j, None)
				try:
					n = int(n)
				except Exception:
					continue
				keys.append(rsa.RSAPublicNumbers(e, n))

			for j in ['CTR2_DEV', 'CTR2_PROD']:
				n = p9_asym.get(j, None)
				try:
					x = int(n['X'])
					y = int(n['Y'])
				except Exception:
					continue
				cert_keys[j] = ec.EllipticCurvePublicNumbers(x, y, ec.SECT233R1())

			for j in ['IVS_CMAC_KEYY', 'IVS_KEYY']:
				k = p9_aes.get(j, None)
				try:
					k = int(k)
				except Exception:
					continue
				aes_keys.append(k)

		except Exception:
			continue

	for i in keys:
		_hash = hashlib.sha256(hex(i.n).encode('ascii')).hexdigest()
		name = _p9_key_hashes.get(_hash, None)
		if name is None or name not in ['CDN_TIK_WRAP', 'LFCS_RETAIL', 'ROOT_CERT_DEV', 'ROOT_CERT_RETAIL']:
			continue

		if name == 'LFCS_RETAIL':
			_cfg_pubnums[name] = rsa.RSAPublicNumbers(e, i.n)
		else:
			_p9_pubnums[name] = rsa.RSAPublicNumbers(e, i.n)

	for i in certs:
		cn = [j.rfc4514_string() for j in i.subject.rdns if j.rfc4514_string().startswith('CN=')]
		if len(cn) != 1:
			continue
		cn = cn[0][3:]

		names = ['NintendoCTR2dev', 'NintendoCTR2prod']
		index = names.index(cn) if cn in names else None
		if index is None:
			continue

		pubnums = i.public_key().public_numbers()
		if not isinstance(pubnums.curve, ec.SECT233R1):
			continue

		hashx = hashlib.sha256(hex(pubnums.x).encode('ascii')).hexdigest()
		hashy = hashlib.sha256(hex(pubnums.y).encode('ascii')).hexdigest()

		namex = _p9_key_hashes.get(hashx, '')
		namey = _p9_key_hashes.get(hashy, '')

		if namex != ['CTR2_DEV_X', 'CTR2_PROD_X'][index] or namey != ['CTR2_DEV_Y', 'CTR2_PROD_Y'][index]:
			continue

		name = ['CTR2_DEV', 'CTR2_PROD'][index]

		_p9_pubnums[name] = ec.EllipticCurvePublicNumbers(pubnums.x, pubnums.y, ec.SECT233R1())

	for i, j in cert_keys.items():
		hashx = hashlib.sha256(hex(j.x).encode('ascii')).hexdigest()
		hashy = hashlib.sha256(hex(j.y).encode('ascii')).hexdigest()

		namex = _p9_key_hashes.get(hashx, '')
		namey = _p9_key_hashes.get(hashy, '')

		if namex != (i + "_X") or namey != (i + "_Y"):
			continue

		_p9_pubnums[i] = ec.EllipticCurvePublicNumbers(j.x, j.y, ec.SECT233R1())

	for i in aes_keys:
		_hash = hashlib.sha256(hex(i).encode('ascii')).hexdigest()
		name = _p9_key_hashes.get(_hash, None)
		if name is None or name not in ['IVS_CMAC_KEYY', 'IVS_KEYY']:
			continue
		_p9_keys[name] = i

	if codebin: # save only if potential new keys
		_save_p9_keys()

def _save_p9_keys() -> None:
	paths = find_in_config_dirs('ctr_constants.json', True)

	for i in paths:
		try:
			with open(i, 'r+', encoding='utf-8') as j, _constants_json_lock:
				try:
					data = json.load(j)
				except json.decoder.JSONDecodeError:
					data = {}
				j.seek(0)
				p9 = data.get('P9', {})
				p9_asym = p9.get('Asymmetric', {})
				p9_aes = p9.get('AES', {})
				cw = _p9_pubnums.get('CDN_TIK_WRAP', None)
				rd = _p9_pubnums.get('ROOT_CERT_DEV', None)
				rr = _p9_pubnums.get('ROOT_CERT_RETAIL', None)
				c2d = _p9_pubnums.get('CTR2_DEV', None)
				c2p = _p9_pubnums.get('CTR2_PROD', None)
				ivscmac = _p9_keys.get('IVS_CMAC_KEYY', None)
				ivs = _p9_keys.get('IVS_KEYY', None)
				p9_asym['CDN_TIK_WRAP_N'] = f"{cw.n}" if cw is not None else p9_asym.get('CDN_TIK_WRAP_N', None)
				p9_asym['ROOT_CERT_DEV_N'] = f"{rd.n}" if rd is not None else p9_asym.get('ROOT_CERT_DEV_N', None)
				p9_asym['ROOT_CERT_RETAIL_N'] = f"{rr.n}" if rr is not None else p9_asym.get('ROOT_CERT_RETAIL_N', None)
				p9_asym['CTR2_DEV'] = {'X': f"{c2d.x}", 'Y': f"{c2d.y}"} if c2d is not None else p9_asym.get('CTR2_DEV', None)
				p9_asym['CTR2_PROD'] = {'X': f"{c2p.x}", 'Y': f"{c2p.y}"} if c2p is not None else p9_asym.get('CTR2_PROD', None)
				p9_aes['IVS_CMAC_KEYY'] = f"{ivscmac}" if ivscmac is not None else p9_asym.get('IVS_CMAC_KEYY', None)
				p9_aes['IVS_KEYY'] = f"{ivs}" if ivs is not None else p9_asym.get('IVS_KEYY', None)
				p9['Asymmetric'] = p9_asym
				p9['AES'] = p9_aes
				data['P9'] = p9
				out = json.dumps(data, indent=2)
				j.truncate()
				j.write(out)

		except Exception:
			continue

# HW AES Constant

_aes_const_c = None

def _load_const_c(constant: typing.Optional[bytes] = None) -> None:
	global _aes_const_c

	_hash_c = '6e637c42711e65cc28db9fe45d7ef3dbc6952a63b8310407befdd7cd6616286a'

	constants = []

	paths = find_in_config_dirs('ctr_constants.json')

	for i in paths:
		try:
			with open(i, 'r', encoding='utf-8') as j, _constants_json_lock:
				data = json.load(j)
			cfg = data.get('HW', {})
			n = cfg.get('AES_CONST_C', None)
			try:
				n = int(n)
			except Exception:
				continue
			constants.append(n)
		except Exception:
			continue

	for i in constants:
		if hashlib.sha256(str(i).encode('ascii')).hexdigest() != _hash_c:
			continue
		_aes_const_c = i
		return

	paths = find_in_config_dirs('aeshw_keygen_constant')

	for i in paths:
		try:
			with open(i, 'rb') as j:
				key = j.read(16)
			if len(key) != 16:
				continue
			constants.append(int.from_bytes(key, 'big'))
		except Exception:
			continue

	if constant is not None and len(constant) == 16:
		constants.append(int.from_bytes(constant, 'big'))

	for i in constants:
		if hashlib.sha256(str(i).encode('ascii')).hexdigest() != _hash_c:
			continue
		_aes_const_c = i

	if _aes_const_c is not None:
		_save_const_c()

def _save_const_c() -> None:
	paths = find_in_config_dirs('ctr_constants.json', True)

	for i in paths:
		try:
			with open(i, 'r+', encoding='utf-8') as j, _constants_json_lock:
				try:
					data = json.load(j)
				except json.decoder.JSONDecodeError:
					data = {}
				j.seek(0)
				hw = data.get('HW', {})
				hw['AES_CONST_C'] = f"{_aes_const_c}" if _aes_const_c is not None else None
				data['HW'] = hw
				out = json.dumps(data, indent=2)
				j.truncate()
				j.write(out)

		except Exception:
			continue

# SSL

def _configure_certs(codebin: bytes) -> None:
	off = -1
	certs = []

	while True:
		_off = codebin[off+1:].find(b'\x30\x82')
		if _off == -1:
			break
		off = off + 1 + _off
		try:
			size = unpack(">2xH", codebin[off:off+4])[0]
			cert = x509.load_der_x509_certificate(codebin[off:off+size+4])
			if isinstance(cert, x509.Certificate):
				certs.append(cert)
		except Exception:
			pass

	bundle_paths = find_in_config_dirs('SSLCertificates/3ds-ssl-ca-bundle.crt', True)
	bundle_crts = []

	for i in bundle_paths:
		try:
			bundle_crts.append([open(i, 'wb'), i])
		except Exception:
			pass

	for i in certs:
		cn = [j.rfc4514_string() for j in i.subject.rdns if j.rfc4514_string().startswith('CN=')]
		if len(cn) != 1:
			continue
		cn = cn[0]
		name = '_'.join('_'.join(cn[3:].split('/')).split('\\'))
		paths = find_in_config_dirs('SSLCertificates/' + name + '.pem', True)

		data = b"# " + cn[3:].encode('utf-8') + b'\n' + i.public_bytes(serialization.Encoding.PEM) + b'\n'
		for j in paths:
			try:
				with open(j, 'wb') as o:
					o.write(data)
			except Exception:
				try:
					os.unlink(j)
				except Exception:
					pass
				pass

		for j in bundle_crts:
			try:
				j[0].write(data)
			except Exception:
				try:
					j[0].close()
					os.unlink(j[1])
				except Exception:
					pass
				pass

def _configure_cli_cert_and_key(enc_cert: bytes, enc_key: bytes):
	from . import keys # keys imports constants, so I can't import keys on top

	key_retail = keys.get_b9_n(0xD, False)
	key_dev = keys.get_b9_n(0xD, True)

	cert = None
	key = None

	for i in [key_retail, key_dev]:
		try:
			if i is None:
				continue

			aes_key = i.to_bytes(16, 'big')

			decryptor = crypto_ciphers.Cipher(
				crypto_ciphers.algorithms.AES(aes_key),
				crypto_ciphers.modes.CBC(enc_cert[:16]),
				default_backend()
			).decryptor()
			_cert = decryptor.update(enc_cert[16:]) + decryptor.finalize()

			if _cert[:2] != b'\x30\x82':
				continue

			size = unpack(">2xH", _cert[:4])[0]
			cert = x509.load_der_x509_certificate(_cert[:size+4])

			decryptor = crypto_ciphers.Cipher(
				crypto_ciphers.algorithms.AES(aes_key),
				crypto_ciphers.modes.CBC(enc_key[:16]),
				default_backend()
			).decryptor()
			_key = decryptor.update(enc_key[16:]) + decryptor.finalize()

			key = serialization.load_der_private_key(_key, password=None)

			k1 = cert.public_key()
			k2 = key.public_key()

			if type(k1) is type(k2) and k1.public_numbers() == k2.public_numbers():
				break

			cert = None
			key = None
		except Exception:
			pass

	if cert is None or key is None:
		return

	cn = [i.rfc4514_string() for i in cert.subject.rdns if i.rfc4514_string().startswith('CN=')]
	if len(cn) != 1:
		return
	cn = cn[0][3:]

	names = ['CTR Common Prod 1', 'CTR Common Dev 1']
	i = names.index(cn) if cn in names else None

	if i is None:
		return

	names = (
		('ctr-common-1-cert-prod.pem', 'ctr-common-1-key-prod.pem'),
		('ctr-common-1-cert-dev.pem', 'ctr-common-1-key-dev.pem')
	)[i]

	for i in find_in_config_dirs('SSLCertificates/client/' + names[0], True):
		try:
			with open(i, 'wb') as o:
				o.write(cert.public_bytes(serialization.Encoding.PEM))
		except Exception:
			try:
				os.unlink(i)
			except Exception:
				pass
			pass

	for i in find_in_config_dirs('SSLCertificates/client/' + names[1], True):
		try:
			with open(i, 'wb') as o:
				o.write(key.private_bytes(
					encoding=serialization.Encoding.PEM,
					format=serialization.PrivateFormat.TraditionalOpenSSL,
					encryption_algorithm=serialization.NoEncryption()
				))
		except Exception:
			try:
				os.unlink(i)
			except Exception:
				pass
			pass



from enum import IntEnum, unique
from struct import unpack
import typing, hashlib

from cryptography.hazmat.primitives.asymmetric import ec, rsa

from . import digitalsignature as digsign
from .exception import CTRExceptionBase, ClassInitError, DataProcessingError

__all__ = [
	"TMD",
	"TitleMetaData",
	"ContentTypeFlags",
	"ContentInfoRecord",
	"ContentChunkRecord"
]

@unique
class ContentTypeFlags(IntEnum):
	Encrypted = 0x1
	Disc = 0x2
	CFM = 0x4
	Optional = 0x4000
	Shared = 0x8000

class ContentInfoRecord:
	def __init__(self, data: typing.SupportsBytes):
		data = bytes(data)[:0x24]

		if len(data) < 0x24:
			raise ClassInitError("Content Info Record does not have sufficient data!")

		segments = unpack(">HH32s", data)

		self._content_index_offset = segments[0]
		self._content_count = segments[1]
		self._content_chunks_sha256 = segments[2]

		self._block = data

	def __bytes__(self):
		return self._block

	def __len__(self):
		return len(self._block)

	def __repr__(self):
		modname = self.__class__.__module__
		clsname = self.__class__.__qualname__
		return f"<{modname}.{clsname} index_offset={self.index_offset} count={self.count} sha256={self.sha256.hex()}>"

	@property
	def index_offset(self) -> int:
		return self._content_index_offset

	@property
	def count(self) -> int:
		return self._content_count

	@property
	def sha256(self) -> bytes:
		return self._content_chunks_sha256

class ContentChunkRecord:
	def __init__(self, data: typing.SupportsBytes):
		data = bytes(data)[:0x30]

		if len(data) < 0x30:
			raise ClassInitError("Content Chunk Record does not have sufficient data!")

		segments = unpack(">IHHQ32s", data)

		self._content_id = segments[0]
		self._content_index = segments[1]
		self._content_type = segments[2]
		self._content_size = segments[3]
		self._content_sha256 = segments[4]

		self._block = data

	def __bytes__(self):
		return self._block

	def __len__(self):
		return len(self._block)

	def __repr__(self):
		modname = self.__class__.__module__
		clsname = self.__class__.__qualname__
		return f"<{modname}.{clsname} id={self.id} index={self.index} type={self.type} size={self.size} sha256={self.sha256.hex()}>"

	@property
	def iv(self) -> bytes:
		return int.to_bytes(self._content_index, 'big').ljust(16, b'\x00')

	@property
	def id(self) -> int:
		return self._content_id

	@property
	def index(self) -> int:
		return self._content_index

	@property
	def type(self) -> int:
		return self._content_type

	@property
	def size(self) -> int:
		return self._content_size

	@property
	def sha256(self) -> bytes:
		return self._content_sha256

class TMD(digsign.IssuedObject):
	def __init__(self, data: typing.SupportsBytes):
		data = bytes(data)

		self._sign = digsign.DigitalSignature(data)

		data = data[len(self._sign):]

		if len(data) < 0x9C4: # 0x9C4 is TMD block size without content chunk records
			raise ClassInitError("Not enough data for TMD")

		segments = unpack(">64s BBB 1s QQ I H 8s 4s B 49s I HHH 2x 32s", data[:0xC4])

		self._issuer_str = segments[0].decode('ascii','replace').split('\x00', 1)[0]
		self._version = segments[1]
		self._ca_crl_version = segments[2]
		self._signed_crl_version = segments[3]
		self._reserved1 = segments[4]
		self._system_version = segments[5]
		self._title_id = segments[6]
		self._title_type = segments[7]
		self._group_id = segments[8]
		self._save_data_size, self._srl_private_save_data_size = unpack("<II", segments[9])
		self._reserved2 = segments[10]
		self._srl_flag = segments[11]
		self._reserved3 = segments[12]
		self._access_rights = segments[13]
		self._title_version = segments[14]
		self._content_count = segments[15]
		self._boot_content = segments[16]
		self._sha256_content_info_records = segments[17]

		if not len(self._issuer_str):
			raise ClassInitError("No TMD issuer")

		if self._issuer_str.find('\uFFFD') != -1:
			raise ClassInitError("Invalid TMD issuer")

		if self._version != 1:
			raise ClassInitError("Unexpected TMD version")

		if not self._content_count:
			raise ClassInitError("No TMD contents")

		if len(data) < 0x9C4 + 0x30 * self._content_count:
			raise ClassInitError("Not enough data for TMD")

		if hashlib.sha256(data[0xC4:0x9C4]).digest() != self._sha256_content_info_records:
			raise ClassInitError("Invalid SHA256 for content infos")

		self._tmd_block = data[:0x9C4 + 0x30 * self._content_count]

		self._content_info_records = tuple([
			ContentInfoRecord(data[i:i+0x24]) for i in range(0xC4, 0x9C4, 0x24)
		])

		effective_info_records = [i for i in self._content_info_records if i.count]
		if sum([i.count for i in effective_info_records]) != self._content_count:
			raise ClassInitError("Content info records don't validate same amount of content chunk records")

		index_list = [i.index_offset for i in effective_info_records]
		if index_list != sorted(index_list) or len(set(index_list)) != len(index_list):
			raise ClassInitError("Content info records has unsorted or repeated indexes")

		self._content_chunk_records = tuple([
			ContentChunkRecord(data[i:i+0x30]) for i in range(0x9C4, 0x9C4 + 0x30 * self._content_count, 0x30)
		])

		chunks = list(self._content_chunk_records)
		for i in effective_info_records:
			sha256 = hashlib.sha256(b''.join([bytes(j) for j in chunks[:i.count]])).digest()
			if sha256 != i.sha256:
				raise ClassInitError("Invalid content chunk data, sha256 mismatch")

			chunks = chunks[i.count:]

		# with enough cursed combinations it may allow repeated indexes.. but I don't wanna test it yet
		# cant test with CIA files this one
		index_list = [i.index for i in self._content_chunk_records]
		if len(set(index_list)) != len(index_list):
			raise ClassInitError("Content chunk records has repeated indexes")

		content_ids = {}
		for i in self._content_chunk_records:
			j = content_ids.get(i.id, [])
			j.append(i)
			content_ids[i.id] = j

		# because of course, content ids can be repeated, for some cursed reason
		for k in (j for i,j in content_ids.items() if len(j) > 1):
			if len(set(((l.sha256, l.size) for l in k))) > 1:
				raise ClassInitError("Repeated content ids with different size or hash detected")

		super().__init__(
			lambda: self._tmd_block[:0xC4],
			lambda: self._sign,
			lambda: self.issuer,
			lambda issuer: issuer.ljust(64,'\x00').encode() + self._tmd_block[0x40:0xC4],
			self._set_reissued
		)

	def __bytes__(self):
		return bytes(self._sign) + self._tmd_block

	def __len__(self):
		return len(self._sign) + len(self._tmd_block)

	def _set_reissued(self, sign: digsign.DigitalSignature, issuer: str, data: typing.SupportsBytes) -> None:
		self._sign = sign
		self._issuer_str = issuer
		self._tmd_block = bytes(data) + self._tmd_block[0xC4:]

	@property
	def issuer(self) -> str:
		return self._issuer_str

	@property
	def version(self) -> int:
		return self._version

	@property
	def ca_crl_version(self) -> int:
		return self._ca_crl_version

	@property
	def signed_crl_version(self) -> int:
		return self._signed_crl_version

	@property
	def system_version(self) -> int:
		return self._system_version

	@property
	def title_id(self) -> int:
		return self._title_id

	@property
	def title_type(self) -> int:
		return self._title_type

	@property
	def group_id(self) -> int:
		return self._group_id

	@property
	def save_data_size(self) -> int:
		return self._save_data_size

	@property
	def srl_private_save_data_size(self) -> int:
		return self._srl_private_save_data_size

	@property
	def srl_flag(self) -> int:
		return self._srl_flag

	@property
	def access_rights(self) -> int:
		return self._access_rights

	@property
	def title_version(self) -> int:
		return self._title_version

	@property
	def boot_content(self) -> int:
		return self._boot_content

	@property
	def chunk_info_records_sha256(self) -> bytes:
		return self._sha256_content_info_records

	@property
	def content_info_records(self) -> typing.Iterable[ContentInfoRecord]:
		return self._content_info_records

	@property
	def content_chunk_records(self) -> typing.Iterable[ContentChunkRecord]:
		return self._content_chunk_records

TitleMetaData = TMD



__version__ = '0.1.3'
__author__ = 'Luis Marques'
__license__ = 'Unlicense'

from . import exception

from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives.asymmetric import ec

if not default_backend().elliptic_curve_supported(ec.SECT233R1()):
	raise exception.ModuleInitError("cryptography's default backend does not support sect233r1!")

del ec
del default_backend
del exception



import typing

from ...nintendowifi import soapenvelopebase

# not for direct use with _xml_xxx_parse
def _parse_attribute_member_impl(
	parent: soapenvelopebase.SoapEnvelopeBase,
	element: soapenvelopebase.XML_Element,
	attr_prefix: bool
) -> typing.Tuple[str,str]:
	parent._xml_raise_if_text(element)
	name = parent._xml_element_parse(element, 'urn:AttributeName' if attr_prefix else 'urn:Name', parent._xml_get_str_element)
	value = parent._xml_element_parse(element, 'urn:AttributeValue' if attr_prefix else 'urn:Value', parent._xml_get_str_element)
	return (name, value)

def _parse_attribute_member_prefixed(
	parent: soapenvelopebase.SoapEnvelopeBase,
	element: soapenvelopebase.XML_Element
) -> typing.Tuple[str,str]:
	return _parse_attribute_member_impl(parent, element, True)

def _parse_attribute_member(
	parent: soapenvelopebase.SoapEnvelopeBase,
	element: soapenvelopebase.XML_Element
) -> typing.Tuple[str,str]:
	return _parse_attribute_member_impl(parent, element, False)



from struct import pack, unpack
import hashlib, typing

from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives.asymmetric import ec

from . import certificate, keys
from . import digitalsignature as digsign
from .exception import ClassInitError
from .constants import _load_p9_keys, _p9_pubnums

__all__ = [
	"OTP",
	"CTCert"
]

_load_p9_keys()

class OTP:
	def __init__(self, data: typing.SupportsBytes):
		data = bytes(data)[:0x100]
		orig_data = data

		if len(data) != 0x100:
			raise ClassInitError("Need 256 bytes for full OTP")

		segments = unpack("<II16sBB6s4s32s60s16s80s32s", data)

		if segments[0] != 0xDEADB00F:
			data = keys._decrypt_otp(orig_data, False)
			if data:
				segments = unpack("<II16sBB6s4s32s60s16s80s32s", data)

		if segments[0] != 0xDEADB00F:
			data = keys._decrypt_otp(orig_data, True)
			if data:
				segments = unpack("<II16sBB6s4s32s60s16s80s32s", data)

		if segments[0] != 0xDEADB00F:
			raise ClassInitError("OTP invalid start")

		self._otp_dec_data = data if data else orig_data

		self._device_id = segments[1]
		self._base_movable_keyY = segments[2]
		self._otp_version = segments[3]
		self._otp_system_type = segments[4]
		self._soc_date = segments[5]
		self._ct_cert_expiration = segments[6]
		self._ct_cert_privk = int.from_bytes(segments[7], 'big') & ((1 << 233) - 1)
		self._ct_cert_signature = segments[8]
		self._hash = segments[11]

		self._lfcs_id, self._supplementary_id = unpack("<QQ", self._base_movable_keyY)

		if hashlib.sha256(self._otp_dec_data[:0xE0]).digest() != self._hash:
			raise ClassInitError("OTP SHA256 verification failed")

		self._otp_enc_data = keys._encrypt_otp(self._otp_dec_data, self.is_dev)
		self._console_keys_x = keys._gen_console_keys(self._otp_dec_data[0x90:0xAC], self.is_dev)
		self._console_keys_x = self._console_keys_x if self._console_keys_x else (None,)*64

	def __bytes__(self):
		return self._otp_dec_data

	def __len__(self):
		return len(self._otp_dec_data)

	def gen_console_keys(self) -> bool:
		if self._console_keys_x.count(None) == 30:
			return True
		self._console_keys_x = keys._gen_console_keys(self._otp_dec_data[0x90:0xAC], self.is_dev)
		self._console_keys_x = self._console_keys_x if self._console_keys_x else (None,)*64
		return self._console_keys_x.count(None) == 30

	@property
	def otp_dec(self) -> bytes:
		return self._otp_dec_data

	@property
	def otp_enc(self) -> bytes:
		return self._otp_enc_data

	@property
	def is_dev(self) -> bool:
		return self._otp_system_type != 0

	@property
	def is_n3ds(self) -> bool:
		return (self._device_id & 0x80000000) != 0

	@property
	def device_id(self) -> int:
		return self._device_id

	@property
	def otp_system_type(self) -> int:
		return self._otp_system_type

	@property
	def otp_version(self) -> int:
		return self._otp_version

	@property
	def lfcs_id(self) -> int:
		return self._lfcs_id

	@property
	def supplementary_id(self) -> int:
		return self._supplementary_id

	@property
	def console_keys_x(self) -> typing.Iterable[typing.Optional[int]]:
		return self._console_keys_x

class CTCert(certificate.Certificate):
	def __init__(self, otp: OTP):
		if not isinstance(otp, OTP):
			raise ClassInitError("CTCert excepts OTP object")

		try:
			privkey = ec.derive_private_key(otp._ct_cert_privk, ec.SECT233R1(), default_backend())
			pubkeynumbers = privkey.public_key().public_numbers()
		except Exception as e:
			raise ClassInitError("EC Key derivation error") from e

		try:
			data = pack(
				">I60s64x64sI64s4s30s30s60x",
				digsign.SignatureType.ECC_SHA256,
				otp._ct_cert_signature,
				(b"Nintendo CA - G3_NintendoCTR2" + (b"prod" if otp._otp_system_type == 0 else b"dev")),
				digsign.KeyType.ECC,
				(f"CT{otp._device_id:08X}-{otp._otp_system_type:02X}").encode(),
				(otp._ct_cert_expiration if otp._otp_version < 5 else otp._ct_cert_expiration[::-1]),
				pubkeynumbers.x.to_bytes(30, 'big'),
				pubkeynumbers.y.to_bytes(30, 'big'),
			)
		except Exception as e:
			raise ClassInitError("Packing error") from e

		super().__init__(data)

		self._is_dev = otp._otp_system_type != 0
		self._device_id = otp._device_id
		self._otp_system_type = otp._otp_system_type
		self._otp_version = otp._otp_version

		try:
			key = _p9_pubnums['CTR2_DEV'] if self._is_dev else _p9_pubnums['CTR2_PROD']
			if key is None:
				_load_p9_keys()
				key = _p9_pubnums['CTR2_DEV'] if self._is_dev else _p9_pubnums['CTR2_PROD']
			key = key.public_key(default_backend())
			self.verify(key)
		except Exception as e:
			raise ClassInitError("Could not verify CTCert") from e

		if not self.set_private_key(privkey.private_numbers()):
			raise ClassInitError("CTCert private key was not successfully loaded!")

	@property
	def is_dev(self) -> bool:
		return self._is_dev

	@property
	def is_retail(self) -> bool:
		return not self._is_dev

	@property
	def device_id(self) -> int:
		return self._device_id

	@property
	def otp_system_type(self) -> int:
		return self._otp_system_type

	@property
	def otp_version(self) -> int:
		return self._otp_version



from struct import pack, unpack
import typing

from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives.asymmetric import ec, rsa

from . import digitalsignature as digsign
from .constants import _load_p9_keys, _p9_pubnums
from .exception import CTRExceptionBase, ClassInitError

__all__ = [
	"Certificate",
	"CertificateKeyType",
	"GetRootCertKey",
	"CertificateKeyTypeError"
]

_load_p9_keys()

class CertificateKeyTypeError(digsign.KeyTypeError):
	"""General CertificateKeyType exception"""

class CertificateKeyType:
	@staticmethod
	def pubkey_block_size(entry: typing.Union[digsign.KeyType, int]) -> int:
		value = int(entry)
		if value == digsign.KeyType.RSA_4096:
			return 0x200 + 4
		if value == digsign.KeyType.RSA_2048:
			return 0x100 + 4
		if value == digsign.KeyType.ECC:
			return 0x3C
		raise CertificateKeyTypeError("Invalid KeyType")

	@staticmethod
	def pubkey_padding_block_size(entry: typing.Union[digsign.KeyType, int]) -> int:
		value = int(entry)
		if value == digsign.KeyType.RSA_4096 or value == digsign.KeyType.RSA_2048:
			return 0x34
		if value == digsign.KeyType.ECC:
			return 0x3C
		raise CertificateKeyTypeError("Invalid KeyType")

class Certificate(digsign.IssuedObject, digsign.IssuingObject):
	def __init__(self, data: typing.SupportsBytes):
		data = bytes(data)[:0x500]

		self._sign = digsign.DigitalSignature(data)

		data = data[len(self._sign):]

		if len(data) < 0x100: # 0x100 is lowest certificate block size
			raise ClassInitError("Not enough data for valid certificate")

		segments = unpack(">64sI64sI", data[:0x88])

		self._issuer_str = segments[0].decode('ascii','replace').split('\x00', 1)[0]
		self._key_type = segments[1]
		self._name_str = segments[2].decode('ascii','replace').split('\x00', 1)[0]
		self._expiration_or_unknown = segments[3]

		if not len(self._issuer_str):
			raise ClassInitError("No certificate issuer")

		if self._issuer_str.find('\uFFFD') != -1:
			raise ClassInitError("Invalid certificate issuer")

		if not len(self._name_str):
			raise ClassInitError("No certificate name")

		if self._name_str.find('\uFFFD') != -1:
			raise ClassInitError("Invalid certificate name")

		block_size = CertificateKeyType.pubkey_block_size(self._key_type)
		pad_size = CertificateKeyType.pubkey_padding_block_size(self._key_type)

		self._cert_block = data[:0x88+block_size+pad_size]

		digsign.IssuingObject.__init__(
			self,
			lambda: self.key_type,
			self.public_key,
			self.private_key,
			lambda: self.full_issuer
		)

		digsign.IssuedObject.__init__(
			self,
			lambda: self._cert_block,
			lambda: self._sign,
			lambda: self.issuer,
			lambda issuer: issuer.ljust(64,'\x00').encode() + self._cert_block[0x40:],
			self._set_reissued
		)

	def __bytes__(self):
		return bytes(self._sign) + self._cert_block

	def __len__(self):
		block_size = CertificateKeyType.pubkey_block_size(self._key_type)
		pad_size = CertificateKeyType.pubkey_padding_block_size(self._key_type)

		return len(self._sign) + 0x88 + block_size + pad_size

	def __repr__(self):
		modname = self.__class__.__module__
		clsname = self.__class__.__qualname__
		return f"<{modname}.{clsname} sign_type={repr(self.sign_type)} issuer={self.issuer} name={self.name} key_type={repr(self.key_type)}>"

	def _set_reissued(self, sign: digsign.DigitalSignature, issuer: str, data: typing.SupportsBytes) -> None:
		self._sign = sign
		self._issuer_str = issuer
		self._cert_block = bytes(data)

	@property
	def issuer(self) -> str:
		return self._issuer_str

	@property
	def name(self) -> str:
		return self._name_str

	@property
	def full_issuer(self) -> str:
		return self._issuer_str + '-' + self._name_str

	@property
	def key_type(self) -> digsign.KeyType:
		return digsign.KeyType(self._key_type)

	@property
	def sign_type(self) -> digsign.SignatureType:
		return self._sign.sign_type

	@property
	def cert_block(self) -> bytes:
		return self._cert_block

	def public_key(
		self
	) -> typing.Union[
		ec.EllipticCurvePublicKey,
		ec.EllipticCurvePublicKeyWithSerialization,
		rsa.RSAPublicKey,
		rsa.RSAPublicKeyWithSerialization
	]:
		block_size = CertificateKeyType.pubkey_block_size(self._key_type)

		if block_size == 0x3C:
			x, y = unpack("30s30s", self._cert_block[0x88:0x88+60])
			x = int(x.hex(), 16) & ((1 << 233) - 1)
			y = int(y.hex(), 16) & ((1 << 233) - 1)
			return ec.EllipticCurvePublicNumbers(x, y, ec.SECT233R1()).public_key(default_backend())

		n, e = unpack(f">{block_size - 4}sI", self._cert_block[0x88:0x88+block_size])
		n = int.from_bytes(n, 'big')
		return rsa.RSAPublicNumbers(e, n).public_key(default_backend())

	def private_key(
		self
	) -> typing.Union[
		ec.EllipticCurvePrivateKey,
		ec.EllipticCurvePrivateKeyWithSerialization,
		rsa.RSAPrivateKey,
		rsa.RSAPrivateKeyWithSerialization,
		None
	]:
		try:
			if isinstance(self._privkey_numbers, ec.EllipticCurvePrivateNumbers):
				return ec.derive_private_key(self._privkey_numbers.private_value, ec.SECT233R1(), default_backend())
			elif isinstance(self._privkey_numbers, rsa.RSAPrivateNumbers):
				return self._privkey_numbers.private_key(default_backend())
		except Exception:
			return None

	def set_private_key(
		self,
		private_numbers: typing.Union[ec.EllipticCurvePrivateNumbers, rsa.RSAPrivateNumbers]
	) -> bool:
		block_size = CertificateKeyType.pubkey_block_size(self._key_type)

		try:
			pubkey_numbers = self.public_key().public_numbers()
			if isinstance(private_numbers, ec.EllipticCurvePrivateNumbers):
				# .private_key() fails?? but derive key is just fine??? idk man
				privkey_test = ec.derive_private_key(private_numbers.private_value, ec.SECT233R1(), default_backend())
			elif isinstance(private_numbers, rsa.RSAPrivateNumbers):
				privkey_test = private_numbers.private_key(default_backend())
			else:
				raise Exception("")
			pubkey_test_numbers = privkey_test.public_key().public_numbers()
		except Exception as e:
			return False

		if block_size == 0x3C:
			if pubkey_numbers.x != pubkey_test_numbers.x or pubkey_numbers.y != pubkey_test_numbers.y:
				return False
		else:
			if pubkey_numbers.n != pubkey_test_numbers.n or pubkey_numbers.e != pubkey_test_numbers.e:
				return False

		self._privkey_numbers = privkey_test.private_numbers()
		return True

	@classmethod
	def process_chain(cls, buffer: typing.SupportsBytes) -> typing.Iterable['Certificate']:
		chain = []

		while (buffer):
			cert = cls(buffer)
			chain.append(cert)
			buffer = buffer[len(cert):]

		return chain

def GetRootCertKey(dev: bool) -> typing.Union[rsa.RSAPublicKey, rsa.RSAPublicKeyWithSerialization, None]:
	key = _p9_pubnums['ROOT_CERT_DEV'] if dev else _p9_pubnums['ROOT_CERT_RETAIL']
	if key is None:
		_load_p9_keys()
		key = _p9_pubnums['ROOT_CERT_DEV'] if dev else _p9_pubnums['ROOT_CERT_RETAIL']
		if key is None:
			return None

	return key.public_key(default_backend())




__all__ = [
	"CleanintyExceptionBase",
	"ModuleInitError",
	"ClassInitError",
	"DataProcessingError"
]

class CleanintyExceptionBase(Exception):
	"""Base exception for all module specific exceptions"""

class ModuleInitError(CleanintyExceptionBase):
	"""Exception related to module initialization"""

class ClassInitError(CleanintyExceptionBase):
	"""Exception related to class initialization"""

class DataProcessingError(CleanintyExceptionBase):
	"""Exception related to error processing data"""




import base64, typing, random

from ...nintendowifi.soapsessionmanager import SoapSessionManager
from ...nintendowifi import soapenvelopebase
from .manager import CtrSoapManager
from ..exception import ClassInitError
from .exception import OperationError, CTRExceptionBase
from ...connection import Connection
from ..ssl import _ssl_certs
from ..title import Title, MediaType
from .._sys_titles import o3ds_sys_titles, n3ds_sys_titles
from .. import regionaldata
from ..ticket import Ticket
from ..certificate import Certificate

# no NUS commands left to add

__all__ = [
	"CtrStubManager",
	"GetSystemTitleHash",
	"GetSystemUpdate",
	"GetSystemCommonETicket",
	"StubCallError"
]

class StubCallError(CTRExceptionBase):
	"""Stub called function"""

class CtrStubManager:
	def __init__(
		self,
		is_n3ds: bool,
		randomize: bool = False,
		region: regionaldata.RegionType = regionaldata.Region.JPN,
		is_dev: bool = False
	):
		self._device_id = (4 << 32) | (random.randint(0, 0x7FFFFFFF) if randomize else 0) | (0x80000000 if is_n3ds else 0)

		self._region = regionaldata.Region.get_region(region)
		if self._region is None:
			self._region = regionaldata.Region.JPN
		elif self._region == regionaldata.Region.AUS:
			self._region = regionaldata.Region.EUR
		self._country = random.choice(
			regionaldata.Country.get_region_country_list(self._region)
		) if randomize else (
			regionaldata.Country.JP,
			regionaldata.Country.US,
			regionaldata.Country.GB,
			None,
			regionaldata.Country.CN,
			regionaldata.Country.KR,
			regionaldata.Country.TW
		)[self._region]
		self._language = regionaldata.Language.get_region_language_list(self._region)[0]

		self._sys_titles = (n3ds_sys_titles if is_n3ds else o3ds_sys_titles)[self._region]
		self._sys_titles = tuple() if self._sys_titles is None else self._sys_titles

		self._serial = random.choice(('Y', 'Q', 'N') if is_n3ds else ('C', 'S', 'A')) if randomize else ('Y' if is_n3ds else 'C')
		self._serial += ('JF', 'W', 'EH', None, 'CF', 'KF', 'TF')[self._region]
		self._serial += f'{(random.choice((1,4,7,9)) * 10000000) + random.randint(0, 9999999)}' if randomize else '10000001'

		self._region = self._region.get_region_str(self._region)
		self._country = self._country.get_country_str(self._country)
		self._language = self._language.get_language_str(self._language)

		self._is_dev = bool(is_dev)

		self._connection = Connection()

	def sleep_with_rng_curve(
		self,
		sleep_min_time: typing.Union[int, float],
		rng_multiplier: typing.Union[int, float] = 1
	):
		pass

	def get_connection(self) -> Connection:
		return self._connection

	def set_service_standby(self, status: bool) -> None:
		raise StubCallError("standby mode is an ecs operation, manager is a nus stub")

	@property
	def is_time_emu_enabled(self) -> bool:
		return False

	@property
	def service_standby(self) -> bool:
		raise StubCallError("standby mode is an ecs operation, manager is a nus stub")

	@property
	def device_id(self) -> int:
		return self._device_id

	@property
	def message_id_step(self) -> int:
		return 0

	@property
	def has_tokens(self) -> bool:
		raise StubCallError("tokens don't apply on nus, manager is a nus stub")

	@property
	def st_token(self) -> typing.Optional[str]:
		raise StubCallError("tokens don't apply on nus, manager is a nus stub")

	@property
	def wt_token(self) -> typing.Optional[str]:
		raise StubCallError("tokens don't apply on nus, manager is a nus stub")

	@property
	def account_id(self) -> typing.Optional[int]:
		raise StubCallError("account ids don't apply on nus, manager is a nus stub")

	@property
	def account_status(self) -> typing.Optional[str]:
		raise StubCallError("account status don't apply on nus, manager is a nus stub")

	@property
	def application_id(self) -> typing.Optional[int]:
		raise StubCallError("application id applies to ecs and cas, manager is a nus stub")

	@property
	def tin(self) -> typing.Optional[int]:
		raise StubCallError("tin applies to ecs and cas, manager is a nus stub")

	@property
	def age(self) -> typing.Optional[int]:
		raise StubCallError("age doesn't apply to nus, manager is a nus stub")

	@property
	def region(self) -> typing.Optional[str]:
		return self._region

	@property
	def country(self) -> typing.Optional[str]:
		return self._country

	@property
	def language(self) -> typing.Optional[str]:
		return self._language

	@property
	def serial_no(self) -> str:
		return self._serial

	@property
	def user_agent(self) -> str:
		return 'CTR NUP 040600 Mar 14 2012 13:32:39'

	@property
	def keepalive(self) -> bool:
		return True

	@property
	def ssl_cert_path(self) -> typing.Optional[str]:
		return _ssl_certs._ca_id_path(3)

	@property
	def ssl_cli_cert_paths(self) -> typing.Optional[typing.Tuple[str, ...]]:
		return _ssl_certs._client_cert_path_tuple(self._is_dev)

	def titles(self, media: MediaType) -> typing.Iterable[Title]:
		return self._sys_titles if media == MediaType.NAND else tuple()

	def get_url_by_identifier(self, name: str) -> typing.Optional[str]:
		return 'https://nus.c.shop.nintendowifi.net/nus/services/NetUpdateSOAP' if name == 'nus' else None

SoapSessionManager.register(CtrStubManager)

def _xml_get_titleversion_element(
	parent: soapenvelopebase.SoapEnvelopeBase,
	element: soapenvelopebase.XML_Element
) -> typing.Tuple[
	int, int, typing.Optional[int], typing.Optional[int], typing.Optional[int]
]:
	parent._xml_raise_if_text(element)
	titleid = parent._xml_element_parse(element, 'urn:TitleId', parent._xml_get_int_base16_element)
	version = parent._xml_element_parse(element, 'urn:Version', parent._xml_get_int_element)
	fssize = parent._xml_element_parse(element, 'urn:FsSize', parent._xml_get_int_element, True)
	tiksize = parent._xml_element_parse(element, 'urn:TicketSize', parent._xml_get_int_element, True)
	tmdsize = parent._xml_element_parse(element, 'urn:TMDSize', parent._xml_get_int_element, True)
	return (titleid, version, fssize, tiksize, tmdsize)

class GetSystemTitleHash(soapenvelopebase.SoapEnvelopeBase):
	@soapenvelopebase.ObjectTimingEmuHelper(0.9, 0.125)
	def __init__(self, ctrsoapmanager: typing.Union[CtrSoapManager, CtrStubManager]):
		if not isinstance(ctrsoapmanager, CtrSoapManager) and not isinstance(ctrsoapmanager, CtrStubManager):
			raise ClassInitError("Expected CtrSoapManager or CtrStubManager")

		super().__init__(soapenvelopebase.SoapSubNames.NUS, 'GetSystemTitleHash', ctrsoapmanager, False, True, False)

		ret = self._send(ctrsoapmanager.get_url_by_identifier('nus'))
		if ret != 200:
			raise OperationError("Bad HTTP response or connection error, ret = {0}".format(ret))

		response_parse = self._initiate_response_parse()
		self._validate_errorcode(self.errorcode, self.errormessage)
		try:
			response = response_parse[1]
			self._titlehash = self._xml_element_parse(response, 'urn:TitleHash', self._xml_get_str_element)
			if len(self._titlehash) != 32:
				raise soapenvelopebase.XMLParseError("length of TitleHash != 32 bytes")
		except CTRExceptionBase:
			raise
		except Exception as e:
			raise soapenvelopebase.XMLParseError("Unexpected exception while parsing XML") from e

	@property
	def titlehash(self) -> str:
		return self._titlehash

class GetSystemUpdate(soapenvelopebase.SoapEnvelopeBase):
	@soapenvelopebase.ObjectTimingEmuHelper(0.9, 0.125)
	def __init__(self, ctrsoapmanager: typing.Union[CtrSoapManager, CtrStubManager]):
		if not isinstance(ctrsoapmanager, CtrSoapManager) and not isinstance(ctrsoapmanager, CtrStubManager):
			raise ClassInitError("Expected CtrSoapManager or CtrStubManager")

		super().__init__(soapenvelopebase.SoapSubNames.NUS, 'GetSystemUpdate', ctrsoapmanager, False, True, False)

		for i in ctrsoapmanager.titles(MediaType.NAND):
			if not Title.is_ctranysys(i.id):
				continue
			self._push_tag('TitleVersion')
			self._write_tag('TitleId', i.id)
			self._write_tag('Version', i.version)
			self._pop_tag()

		ret = self._send(ctrsoapmanager.get_url_by_identifier('nus'))
		if ret != 200:
			raise OperationError("Bad HTTP response or connection error, ret = {0}".format(ret))

		response_parse = self._initiate_response_parse()
		self._validate_errorcode(self.errorcode, self.errormessage)
		try:
			response = response_parse[1]
			# normal parser only gets title versions and hash, but I get all that I observed appear in a response
			self._contextprefixurl = self._xml_element_parse(response, 'urn:ContentPrefixURL', self._xml_get_str_element, True)
			self._uncachedcontextprefixurl = self._xml_element_parse(response, 'urn:UncachedContentPrefixURL', self._xml_get_str_element, True)
			self._titles = self._xml_multi_element_parse(response, 'urn:TitleVersion', _xml_get_titleversion_element, True)
			self._titles = tuple(self._titles) if self._titles is not None else None
			self._uploadauditdata = self._xml_element_parse(response, 'urn:UploadAuditData', self._xml_get_str_element, True)
			self._titlehash = self._xml_element_parse(response, 'urn:TitleHash', self._xml_get_str_element, self._titles is None)
			if self._titlehash is not None and len(self._titlehash) != 32:
				raise soapenvelopebase.XMLParseError("length of TitleHash != 32 bytes")
		except CTRExceptionBase:
			raise
		except Exception as e:
			raise soapenvelopebase.XMLParseError("Unexpected exception while parsing XML") from e

	@property
	def contextprefixurl(self) -> typing.Optional[str]:
		return self._contextprefixurl

	@property
	def uncachedcontextprefixurl(self) -> typing.Optional[str]:
		return self._uncachedcontextprefixurl

	@property
	def titles(self) -> typing.Iterator[
		typing.Tuple[
			int, int, typing.Optional[int], typing.Optional[int], typing.Optional[int]
		]
	]:
		return self._titles

	@property
	def uploadauditdata(self) -> typing.Optional[str]:
		return self._uploadauditdata

	@property
	def titlehash(self) -> typing.Optional[str]:
		return self._titlehash

class GetSystemCommonETicket(soapenvelopebase.SoapEnvelopeBase):
	@soapenvelopebase.ObjectTimingEmuHelper(0.9, 0.125)
	def __init__(self, ctrsoapmanager: CtrSoapManager, titleids: typing.Iterable[typing.SupportsInt]):
		if not isinstance(ctrsoapmanager, CtrSoapManager):
			raise ClassInitError("Expected CtrSoapManager")

		super().__init__(soapenvelopebase.SoapSubNames.NUS, 'GetSystemCommonETicket', ctrsoapmanager, False, False, False)

		count = 0

		for i in titleids:
			self._write_tag('TitleId', f"{int(i):016X}")
			count += 1

		ret = self._send(ctrsoapmanager.get_url_by_identifier('nus'))
		if ret != 200:
			raise OperationError("Bad HTTP response or connection error, ret = {0}".format(ret))

		response_parse = self._initiate_response_parse()
		self._validate_errorcode(self.errorcode, self.errormessage)

		try:
			response = response_parse[1]
			self._cetks = self._xml_multi_element_parse(response, 'urn:CommonETicket', lambda x,y: Ticket(self._xml_get_base64_element(x, y)), True)
			self._cetks = tuple(self._cetks) if self._cetks is not None else tuple()
			self._certs = self._xml_multi_element_parse(response, 'urn:Certs', lambda x,y: Certificate(self._xml_get_base64_element(x, y)), True)
			self._certs = tuple(self._certs) if self._certs is not None else tuple()
			# title ids requested must be == cetks received, but like ecs:AccountGetETickets, I'll ignore it
			self._got_all = count == len(self._cetks)
		except CTRExceptionBase:
			raise
		except Exception as e:
			raise soapenvelopebase.XMLParseError("Unexpected exception while parsing XML") from e

	@property
	def cetks(self) -> typing.Iterator[Ticket]:
		return self._cetks

	@property
	def certs(self) -> typing.Iterator[Certificate]:
		return self._certs

	@property
	def got_all(self) -> bool:
		return self._got_all



from enum import IntEnum, unique
import typing

from . import regionaldata
from ..exception import ClassInitError

__all__ = [
	"MediaType",
	"Title"
]

@unique
class MediaType(IntEnum):
	NAND = 0
	SD = 1
	GAME_CARD = 2

class Title:
	def __init__(
		self,
		tid: typing.SupportsInt,
		version: typing.SupportsInt,
		region: typing.Optional[regionaldata.RegionType] = None, # None for all
		downloadable: bool = True
	):
		self._tid = int(tid) & 0xFFFFFFFFFFFFFFFF
		self._version = int(version) & 0xFFFF
		self._version_chunks = (
			self._version >> 10,
			(self._version >> 4) & 0x3F,
			self._version & 0xF
		)
		self._region = regionaldata.Region.get_region(region)
		if self._region is None and region is not None:
			raise ClassInitError("Invalid title region")
		self._downloadable = bool(downloadable)

	def __repr__(self):
		modname = self.__class__.__module__
		clsname = self.__class__.__qualname__
		if self._region is None:
			return f"<{modname}.{clsname} id=0x{self._tid:016x} version={self._version} version_chunks={self._version_chunks}>"
		return f"<{modname}.{clsname} id=0x{self._tid:016x} version={self._version} version_chunks={self._version_chunks} region={repr(self._region)}>"

	def __int__(self):
		return self._tid

	def __eq__(self, rhs):
		if isinstance(rhs, int):
			return self._tid == rhs
		if isinstance(rhs, Title):
			return self._tid == rhs._tid and self._version == rhs._version
		return NotImplemented

	def __ne__(self, rhs):
		if isinstance(rhs, int):
			return self._tid != rhs
		if isinstance(rhs, Title):
			return self._tid != rhs._tid or self._version != rhs._version
		return NotImplemented

	def __lt__(self, rhs):
		if isinstance(rhs, int):
			return self._tid < rhs # or (self._tid == rhs and self._version < 0)
		if isinstance(rhs, Title):
			return self._tid < rhs._tid or (self._tid == rhs._tid and self._version < rhs._version)
		return NotImplemented

	def __le__(self, rhs):
		if isinstance(rhs, int):
			return self._tid < rhs or (self._tid == rhs and self._version <= 0)
		if isinstance(rhs, Title):
			return self._tid < rhs._tid or (self._tid == rhs._tid and self._version <= rhs._version)
		return NotImplemented

	def __gt__(self, rhs):
		if isinstance(rhs, int):
			return self._tid > rhs or (self._tid == rhs and self._version > 0)
		if isinstance(rhs, Title):
			return self._tid > rhs._tid or (self._tid == rhs._tid and self._version > rhs._version)
		return NotImplemented

	def __ge__(self, rhs):
		if isinstance(rhs, int):
			return self._tid > rhs or (self._tid == rhs and self._version >= 0)
		if isinstance(rhs, Title):
			return self._tid > rhs._tid or (self._tid == rhs._tid and self._version >= rhs._version)
		return NotImplemented

	@property
	def id(self) -> int:
		return self._tid

	@property
	def version(self) -> int:
		return self._version

	@property
	def version_major(self) -> int:
		return self._version_chunks[0]

	@property
	def version_minor(self) -> int:
		return self._version_chunks[1]

	@property
	def version_revision(self) -> int:
		return self._version_chunks[2]

	@property
	def version_chunks(self) -> typing.Tuple[int, int, int]:
		return self._version_chunks

	@property
	def region(self) -> typing.Optional[regionaldata.Region]:
		return self._region

	@property
	def downloadable(self) -> bool:
		return self._downloadable

	@property
	def export_tuple(self) -> typing.Tuple[int, int, typing.Optional[int], bool]:
		return (self._tid, self._version, int(self._region) if self._region else None, self._downloadable)

	@staticmethod
	def is_ctr(tid: typing.SupportsInt) -> bool:
		return (int(tid) & 0xFFFF400000000000) == 0x0004000000000000

	@staticmethod
	def is_ctrtwl(tid: typing.SupportsInt) -> bool:
		return (int(tid) & 0xFFFFC00000000000) == 0x0004800000000000

	@staticmethod
	def is_ctrsys(tid: typing.SupportsInt) -> bool:
		return (int(tid) & 0xFFFFC01000000000) == 0x0004001000000000

	@staticmethod
	def is_ctrtwlsys(tid: typing.SupportsInt) -> bool:
		return (int(tid) & 0xFFFFC00100000000) == 0x0004800100000000

	@classmethod
	def is_ctranysys(cls, tid: typing.SupportsInt) -> bool:
		return cls.is_ctrsys(tid) or cls.is_ctrtwlsys(tid)

	@staticmethod
	def is_ctrdlc(tid: typing.SupportsInt) -> bool:
		return (int(tid) & 0xFFFFFFFF00000000) == 0x0004008C00000000

	@staticmethod
	def is_ctrdlp(tid: typing.SupportsInt) -> bool:
		return (int(tid) & 0xFFFFFFFF00000000) == 0x0004000100000000

	@staticmethod
	def is_ctrdemo(tid: typing.SupportsInt) -> bool:
		return (int(tid) & 0xFFFFFFFF00000000) == 0x0004000200000000

	@staticmethod
	def is_ctrlicence(tid: typing.SupportsInt) -> bool:
		return (int(tid) & 0xFFFFFFFF00000000) == 0x0004000D00000000

	@staticmethod
	def is_ctrupdate(tid: typing.SupportsInt) -> bool:
		return (int(tid) & 0xFFFFFFFF00000000) == 0x0004000E00000000



from enum import IntEnum, unique
import typing

__all__ = [
	"Region",
	"Country",
	"Language"
]

@unique
class Region(IntEnum):
	JPN = 0
	USA = 1
	EUR = 2
	AUS = 3
	CHN = 4
	KOR = 5
	TWN = 6

	@classmethod
	def get_region(
		cls,
		value: typing.Union['Region', typing.SupportsInt, str]
	) -> typing.Union['Region', None]:
		try:
			if isinstance(value, str):
				return getattr(cls, value[:3].upper())
			else:
				return cls(int(value))
		except:
			return None

	@classmethod
	def get_region_str(
		cls,
		value: typing.Union['Region', typing.SupportsInt, str]
	) -> typing.Union[str, None]:
		region = cls.get_region(value)
		return region.name if region is not None else None

@unique
class Country(IntEnum):
	JP = 1   # Japan
	AI = 8   # Anguilla
	AG = 9   # Antigua and Barbuda
	AR = 10  # Argentina
	AW = 11  # Aruba
	BS = 12  # Bahamas
	BB = 13  # Barbados
	BZ = 14  # Belize
	BO = 15  # Bolivia
	BR = 16  # Brazil
	VG = 17  # British Virgin Islands
	CA = 18  # Canada
	KY = 19  # Cayman Islands
	CL = 20  # Chile
	CO = 21  # Colombia
	CR = 22  # Costa Rica
	DM = 23  # Dominica
	DO = 24  # Dominican Republic
	EC = 25  # Ecuador
	SV = 26  # El Salvador
	GF = 27  # French Guiana
	GD = 28  # Grenada
	GP = 29  # Guadeloupe
	GT = 30  # Guatemala
	GY = 31  # Guyana
	HT = 32  # Haiti
	HN = 33  # Honduras
	JM = 34  # Jamaica
	MQ = 35  # Martinique
	MX = 36  # Mexico
	MS = 37  # Montserrat
	AN = 38  # Netherlands Antilles
	NI = 39  # Nicaragua
	PA = 40  # Panama
	PY = 41  # Paraguay
	PE = 42  # Peru
	KN = 43  # Saint Kitts and Nevis
	LC = 44  # Saint Lucia
	VC = 45  # Saint Vincent and the Grenadines
	SR = 46  # Suriname
	TT = 47  # Trinidad and Tobago
	TC = 48  # Turks and Caicos Islands
	US = 49  # United States
	UY = 50  # Uruguay
	VI = 51  # US Virgin Islands
	VE = 52  # Venezuela
	AL = 64  # Albania
	AU = 65  # Australia
	AT = 66  # Austria
	BE = 67  # Belgium
	BA = 68  # Bosnia and Herzegovina
	BW = 69  # Botswana
	BG = 70  # Bulgaria
	HR = 71  # Croatia
	CY = 72  # Cyprus
	CZ = 73  # Czech Republic
	DK = 74  # Denmark
	EE = 75  # Estonia
	FI = 76  # Finland
	FR = 77  # France
	DE = 78  # Germany
	GR = 79  # Greece
	HU = 80  # Hungary
	IS = 81  # Iceland
	IE = 82  # Ireland
	IT = 83  # Italy
	LV = 84  # Latvia
	LS = 85  # Lesotho
	LI = 86  # Liechtenstein
	LT = 87  # Lithuania
	LU = 88  # Luxembourg
	MK = 89  # Macedonia
	MT = 90  # Malta
	ME = 91  # Montenegro
	MZ = 92  # Mozambique
	NA = 93  # Namibia
	NL = 94  # Netherlands
	NZ = 95  # New Zealand
	NO = 96  # Norway
	PL = 97  # Poland
	PT = 98  # Portugal
	RO = 99  # Romania
	RU = 100 # Russia
	RS = 101 # Serbia and Kosovo
	SK = 102 # Slovakia
	SI = 103 # Slovenia
	ZA = 104 # South Africa
	ES = 105 # Spain
	SZ = 106 # Swaziland
	SE = 107 # Sweden
	CH = 108 # Switzerland
	TR = 109 # Turkey
	GB = 110 # United Kingdom
	ZM = 111 # Zambia
	ZW = 112 # Zimbabwe
	AZ = 113 # Azerbaijan
	MR = 114 # Mauritania
	ML = 115 # Mali
	NE = 116 # Niger
	TD = 117 # Chad
	SD = 118 # Sudan
	ER = 119 # Eritrea
	DJ = 120 # Djibouti
	SO = 121 # Somalia
	AD = 122 # Andorra
	GI = 123 # Gibraltar
	GG = 124 # Guernsey
	IM = 125 # Isle of Man
	JE = 126 # Jersey
	MC = 127 # Monaco
	TW = 128 # Taiwan
	KR = 136 # South Korea
	HK = 144 # Hong Kong
	MO = 145 # Macau
	ID = 152 # Indonesia
	SG = 153 # Singapore
	TH = 154 # Thailand
	PH = 155 # Philippines
	MY = 156 # Malaysia
	CN = 160 # China
	AE = 168 # United Arab Emirates
	IN = 169 # India
	EG = 170 # Egypt
	OM = 171 # Oman
	QA = 172 # Qatar
	KW = 173 # Kuwait
	SA = 174 # Saudi Arabia
	SY = 175 # Syria
	BH = 176 # Bahrain
	JO = 177 # Jordan
	SM = 184 # San Marino
	VA = 185 # Vatican City
	BM = 186 # Bermuda

	@classmethod
	def get_country(
		cls,
		value: typing.Union['Country', typing.SupportsInt, str]
	) -> typing.Union['Country', None]:
		try:
			if isinstance(value, str):
				return getattr(cls, value[:2].upper())
			else:
				return cls(int(value))
		except:
			return None

	@classmethod
	def get_country_str(
		cls,
		value: typing.Union['Country', typing.SupportsInt, str]
	) -> typing.Union[str, None]:
		country = cls.get_country(value)
		return country.name if country is not None else None

	@classmethod
	def get_region_country_list(
		cls,
		region: typing.Union[Region, typing.SupportsInt, str]
	) -> typing.Union[typing.Iterable['Country'], None]:
		region = Region.get_region(region)
		if region is None:
			return None
		if region == Region.JPN:
			return (cls.JP,)
		if region == Region.USA:
			return (cls.AI, cls.AG, cls.AR, cls.AW, cls.BS, cls.BB, cls.BZ, cls.BM, cls.BO, cls.BR, cls.VG, cls.CA, cls.KY, cls.CL, cls.CO, cls.CR, cls.DM, cls.DO, cls.EC, cls.SV, cls.GF, cls.GD, cls.GP, cls.GT, cls.GY, cls.HT, cls.HN, cls.JM, cls.MY, cls.MQ, cls.MX, cls.MS, cls.AN, cls.NI, cls.PA, cls.PY, cls.PE, cls.SA, cls.SG, cls.KN, cls.LC, cls.VC, cls.SR, cls.TT, cls.TC, cls.AE, cls.US, cls.UY, cls.VI, cls.VE)
		if region == Region.EUR:
			return (cls.AL, cls.AD, cls.AU, cls.AT, cls.AZ, cls.BE, cls.BA, cls.BW, cls.BG, cls.TD, cls.HR, cls.CY, cls.CZ, cls.DK, cls.DJ, cls.ER, cls.EE, cls.FI, cls.FR, cls.DE, cls.GI, cls.GR, cls.GG, cls.HU, cls.IS, cls.IN, cls.IE, cls.IM, cls.IT, cls.JE, cls.LV, cls.LS, cls.LI, cls.LT, cls.LU, cls.MK, cls.ML, cls.MT, cls.MR, cls.MC, cls.ME, cls.MZ, cls.NA, cls.NL, cls.NZ, cls.NE, cls.NO, cls.PL, cls.PT, cls.RO, cls.RU, cls.SM, cls.RS, cls.SK, cls.SI, cls.SO, cls.ZA, cls.ES, cls.SD, cls.SZ, cls.SE, cls.CH, cls.TR, cls.GB, cls.VA, cls.ZM, cls.ZW)
		if region == Region.CHN:
			return (cls.CN,)
		if region == Region.KOR:
			return (cls.KR,)
		if region == Region.TWN:
			return (cls.TW, cls.HK)
		return None

	@classmethod
	def get_unknown_region_country_list(cls) -> typing.Iterable['Country']:
		return (cls.MO, cls.ID, cls.TH, cls.PH, cls.EG, cls.OM, cls.QA, cls.KW, cls.SY, cls.BH, cls.JO)

	@classmethod
	def get_region_by_country(
		cls,
		value: typing.Union['Country', typing.SupportsInt, str]
	) -> typing.Union[Region, None]:
		country = cls.get_country(value)
		if country is None:
			return None
		if country in (cls.JP,):
			return Region.JPN
		if country in (cls.AI, cls.AG, cls.AR, cls.AW, cls.BS, cls.BB, cls.BZ, cls.BM, cls.BO, cls.BR, cls.VG, cls.CA, cls.KY, cls.CL, cls.CO, cls.CR, cls.DM, cls.DO, cls.EC, cls.SV, cls.GF, cls.GD, cls.GP, cls.GT, cls.GY, cls.HT, cls.HN, cls.JM, cls.MY, cls.MQ, cls.MX, cls.MS, cls.AN, cls.NI, cls.PA, cls.PY, cls.PE, cls.SA, cls.SG, cls.KN, cls.LC, cls.VC, cls.SR, cls.TT, cls.TC, cls.AE, cls.US, cls.UY, cls.VI, cls.VE):
			return Region.USA
		if country in (cls.AL, cls.AD, cls.AU, cls.AT, cls.AZ, cls.BE, cls.BA, cls.BW, cls.BG, cls.TD, cls.HR, cls.CY, cls.CZ, cls.DK, cls.DJ, cls.ER, cls.EE, cls.FI, cls.FR, cls.DE, cls.GI, cls.GR, cls.GG, cls.HU, cls.IS, cls.IN, cls.IE, cls.IM, cls.IT, cls.JE, cls.LV, cls.LS, cls.LI, cls.LT, cls.LU, cls.MK, cls.ML, cls.MT, cls.MR, cls.MC, cls.ME, cls.MZ, cls.NA, cls.NL, cls.NZ, cls.NE, cls.NO, cls.PL, cls.PT, cls.RO, cls.RU, cls.SM, cls.RS, cls.SK, cls.SI, cls.SO, cls.ZA, cls.ES, cls.SD, cls.SZ, cls.SE, cls.CH, cls.TR, cls.GB, cls.VA, cls.ZM, cls.ZW):
			return Region.EUR
		if country in (cls.CN,):
			return Region.CHN
		if country in (cls.KR,):
			return Region.KOR
		if country in (cls.TW, cls.HK):
			return Region.TWN
		return None

@unique
class Language(IntEnum):
	JA = 0 # japanese
	EN = 1 # english
	FR = 2 # french
	DE = 3 # german
	IT = 4 # italian
	ES = 5 # spanish
	ZH = 6 # simplified chinese
	KO = 7 # korean
	NL = 8 # dutch
	PT = 9 # portuguese
	RU = 10 # russian
	ZH_TRAD = 11 # traditional chinese

	@classmethod
	def get_language(
		cls,
		value: typing.Union['Language', typing.SupportsInt, str]
	) -> typing.Union['Language', None]:
		try:
			if isinstance(value, str):
				return getattr(cls, value[:7].upper())
			else:
				return cls(int(value))
		except:
			return None

	@classmethod
	def get_language_str(
		cls,
		value: typing.Union['Language', typing.SupportsInt, str]
	) -> typing.Union[str, None]:
		language = cls.get_language(value)
		return language.name[:2].lower() if language is not None else None

	@classmethod
	def get_region_language_list(
		cls,
		region: typing.Union[Region, typing.SupportsInt, str]
	) -> typing.Union[typing.Iterable['Language'], None]:
		region = Region.get_region(region)
		if region is None:
			return None
		if region == Region.JPN:
			return (cls.JA,)
		if region == Region.CHN:
			return (cls.ZH,)
		if region == Region.KOR:
			return (cls.KO,)
		if region == Region.TWN:
			return (cls.ZH_TRAD,)
		if region == Region.USA:
			return (cls.EN, cls.FR, cls.ES, cls.PT)
		if region == Region.EUR:
			return (cls.EN, cls.FR, cls.DE, cls.ES, cls.IT, cls.NL, cls.PT, cls.RU)
		return None

RegionType = typing.Union[Region, typing.SupportsInt, str]
CountryType = typing.Union[Country, typing.SupportsInt, str]
LanguageType = typing.Union[Language, typing.SupportsInt, str]



from struct import unpack
import typing, hashlib

from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives.asymmetric import ec, rsa
from cryptography.hazmat.primitives.ciphers import Cipher, algorithms, modes

from . import digitalsignature as digsign
from .exception import CTRExceptionBase, ClassInitError, DataProcessingError
from .otp import CTCert

__all__ = [
	"Ticket"
]

class Ticket(digsign.IssuedObject):
	def __init__(
		self,
		data: typing.SupportsBytes,
		device_cert: typing.Optional[CTCert] = None,
		act_first_import: bool = True # else False, act as a post-import raw dump
	):
		mindata = bytes(data)[:0x3B8]

		self._sign = digsign.DigitalSignature(mindata)

		mindata = mindata[len(self._sign):]

		if len(mindata) < 0x178: # 0x178 is minimal ticket block size
			raise ClassInitError("Not enough data for minimal ticket size")

		segments = unpack(">64s 60s BBB 16s c QIQ 2s H 8s B B 42s I c B 66s 64s 20s", mindata[:0x178])

		self._issuer_str = segments[0].decode('ascii','replace').split('\x00', 1)[0]
		self._ec_pubkey = segments[1]
		self._version = segments[2]
		self._ca_crl_version = segments[3]
		self._signed_crl_version = segments[4]
		self._enc_title_key = segments[5]
		self._reserved1 = segments[6]
		self._ticket_id = segments[7]
		self._device_id = segments[8]
		self._title_id = segments[9]
		self._reserved2 = segments[10]
		self._tik_title_version = segments[11]
		self._reserved3 = segments[12]
		self._license_type = segments[13]
		self._common_index = segments[14]
		self._reserved4 = segments[15]
		self._eshop_id = segments[16]
		self._reserved5 = segments[17]
		self._audit = segments[18]
		self._reserved6 = segments[19]
		self._limits = segments[20]
		self._content_index_header = unpack(">4s II 8s", segments[21])

		if not len(self._issuer_str):
			raise ClassInitError("No ticket issuer")

		if self._issuer_str.find('\uFFFD') != -1:
			raise ClassInitError("Invalid ticket issuer")

		if self._version != 1:
			raise ClassInitError("Unexpected ticket version")

		if self._content_index_header[0] != b'\x00\x01\x00\x14' or \
			self._content_index_header[3] != b'\x00\x01\x00\x14\x00\x00\x00\x00':
			raise ClassInitError("Invalid or unsupported content index")

		if self._content_index_header[1] < 20:
			raise ClassInitError("Invalid content index size")

		self._tik_block_size = 0x164 + self._content_index_header[1]
		self._tik_block = bytes(data)[len(self._sign):len(self._sign) + self._tik_block_size]

		if len(self._tik_block) < self._tik_block_size:
			raise ClassInitError("Insuffient data for full ticket")

		if self._content_index_header[2] < 20 or \
			self._content_index_header[2] < self._content_index_header[1] < self._content_index_header[2] + 20:
			raise ClassInitError("Invalid content index data header")

		if self._content_index_header[1] >= self._content_index_header[2] + 20:
			rights = 0
			data_header_off = 0x164 + self._content_index_header[2]
			data_header = unpack(
				">IIIIH2s",
				self._tik_block[
					data_header_off : data_header_off + 20
				]
			)
			if data_header[4] == 3 and data_header[1] != 0:
				data_off = 0x164 + data_header[0]
				data_off_max = self._tik_block_size
				data_off_end = data_off + data_header[1] * 0x84
				data_off_end = data_off_end if data_off_end < data_off_max else data_off_max
				count = (data_off_end - data_off) // 0x84
				if count <= 0:
					rights = (1 << 256) - 1
				else:
					rights = 0
					last_highest_index = 0
					for i in range(0, count):
						blk_index, blk_rights = unpack(
							">2x H 128s",
							self._tik_block[data_off+0x84*i:data_off+0x84*i+0x84]
						)
						foo = max(last_highest_index - blk_index, 0)
						if foo >= 1024:
							continue
						last_highest_index = max(last_highest_index, blk_index)
						right_block = int.from_bytes(blk_rights, 'little')
						right_block >>= foo
						right_block <<= foo + blk_index
						rights |= right_block
					rights &= (1 << 65536) - 1
			else:
				rights = (1 << 256) - 1

			self._rights = rights
		else:
			self._rights = (1 << 256) - 1 # default rights if not specified

		if device_cert is not None:
			self.fix_and_get_personal_key_tik(device_cert, act_first_import)
		else:
			self._personal_enc_title_key = None

		super().__init__(
			lambda: self._tik_block,
			lambda: self._sign,
			lambda: self.issuer,
			lambda issuer: issuer.ljust(64,'\x00').encode() + self._tik_block[0x40:],
			self._set_reissued
		)

	def __bytes__(self):
		return bytes(self._sign) + self._tik_block

	def __len__(self):
		return len(self._sign) + self._tik_block_size

	def __repr__(self):
		modname = self.__class__.__module__
		clsname = self.__class__.__qualname__
		return f"<{modname}.{clsname} issuer={self.issuer} ticket_id={self.ticket_id} device_id={self.device_id} title_id=0x{self.title_id:016x} eshop_id={self.eshop_id}>"

	def _set_reissued(self, sign: digsign.DigitalSignature, issuer: str, data: typing.SupportsBytes) -> None:
		self._sign = sign
		self._issuer_str = issuer
		self._tik_block = bytes(data)

	@staticmethod
	def _get_personal_tik_key_layer_key_iv(
		tik_id: typing.SupportsInt,
		ec_pubkey_block: typing.SupportsBytes,
		ct_certificate: CTCert
	) -> typing.Tuple[bytes, bytes]:
		if not isinstance(ct_certificate, CTCert):
			raise DataProcessingError("Expected CTCert")

		ct_ec_prikkey = ct_certificate.private_key()
		tik_id = int(tik_id) & ((1 << 64) - 1)
		ec_pubkey_block = bytes(ec_pubkey_block)

		if not ct_ec_prikkey or \
			(not isinstance(ct_ec_prikkey, ec.EllipticCurvePrivateKey) and not isinstance(ct_ec_prikkey, ec.EllipticCurvePrivateKeyWithSerialization)):
			raise DataProcessingError("Invalid ctcert ec key")

		if len(ec_pubkey_block) != 60:
			raise DataProcessingError("Expected 60 byte ec pubkey block")

		x, y = unpack("30s30s", ec_pubkey_block)
		x = int(x.hex(), 16) & ((1 << 233) - 1)
		y = int(y.hex(), 16) & ((1 << 233) - 1)
		tik_pubkey = ec.EllipticCurvePublicNumbers(x, y, ec.SECT233R1()).public_key(default_backend())

		aes_key = hashlib.sha1(ct_ec_prikkey.exchange(ec.ECDH(), tik_pubkey)).digest()[:16]
		iv = int.to_bytes(tik_id << 64, 16, 'big')

		return (aes_key, iv)

	# because one encryption layer was not enough appearantly
	@classmethod
	def _decrypt_personal_tik_key_layer(
		cls,
		enc_enc_title_key: typing.SupportsBytes,
		tik_id: typing.SupportsInt,
		ec_pubkey_block: typing.SupportsBytes,
		ct_certificate: CTCert
	) -> bytes:
		aes_key, iv = cls._get_personal_tik_key_layer_key_iv(tik_id, ec_pubkey_block, ct_certificate)

		enc_enc_title_key = bytes(enc_enc_title_key)

		if len(enc_enc_title_key) != 16:
			raise DataProcessingError("Expected 16 byte encrypted key")

		cipher = Cipher(algorithms.AES(aes_key), modes.CBC(iv))
		decryptor = cipher.decryptor()
		return decryptor.update(enc_enc_title_key) + decryptor.finalize()

	@classmethod
	def _encrypt_personal_tik_key_layer(
		cls,
		dec_enc_title_key: typing.SupportsBytes,
		tik_id: typing.SupportsInt,
		ec_pubkey_block: typing.SupportsBytes,
		ct_certificate: CTCert
	) -> bytes:
		aes_key, iv = cls._get_personal_tik_key_layer_key_iv(tik_id, ec_pubkey_block, ct_certificate)

		dec_enc_title_key = bytes(dec_enc_title_key)

		if len(dec_enc_title_key) != 16:
			raise DataProcessingError("Expected 16 byte decrypted key")

		cipher = Cipher(algorithms.AES(aes_key), modes.CBC(iv))
		encryptor = cipher.encryptor()
		return encryptor.update(dec_enc_title_key) + encryptor.finalize()

	@property
	def issuer(self) -> str:
		return self._issuer_str

	@property
	def ec_pubkey(self) -> bytes:
		return self._ec_pubkey

	@property
	def version(self) -> int:
		return self._version

	@property
	def ca_crl_version(self) -> int:
		return self._ca_crl_version

	@property
	def signed_crl_version(self) -> int:
		return self._signed_crl_version

	@property
	def enc_title_key(self) -> bytes:
		return self._enc_title_key

	@property
	def personal_enc_title_key(self) -> typing.Optional[bytes]:
		return self._personal_enc_title_key

	@property
	def ticket_id(self) -> int:
		return self._ticket_id

	@property
	def device_id(self) -> int:
		return self._device_id

	@property
	def title_id(self) -> int:
		return self._title_id

	@property
	def ticket_title_version(self) -> int:
		return self._tik_title_version

	@property
	def license_type(self) -> int:
		return self._license_type

	@property
	def common_index(self) -> int:
		return self._common_index

	@property
	def eshop_id(self) -> int:
		return self._eshop_id

	@property
	def audit(self) -> int:
		return self._audit

	@property
	def limits(self) -> bytes:
		return self._limits

	@property
	def ticket_block(self) -> bytes:
		return self._tik_block

	@property
	def is_personal(self) -> bool:
		return self._device_id != 0

	@property
	def installable_export(self) -> bytes:
		if not self.is_personal:
			return bytes(self)
		elif self._personal_enc_title_key is None or len(self._personal_enc_title_key) != 16:
			raise DataProcessingError("Missing personal encrypted key!")
		return bytes(self._sign) + self._tik_block[:0x7F] + self._personal_enc_title_key + self._tik_block[0x8F:]

	def fix_and_get_personal_key_tik(
		self,
		device_cert: CTCert,
		act_first_import: bool = True # else False, act as a post-import raw dump
	) -> bool:
		if not self.is_personal:
			self._personal_enc_title_key = None
			return False

		if not isinstance(device_cert, CTCert):
			raise DataProcessingError("Expecting CTCert for device cert if being given")

		if self._device_id != device_cert.device_id:
			raise DataProcessingError("Device certificate's device id must match the one on ticket")

		if act_first_import:
			key = self._decrypt_personal_tik_key_layer(
				self._enc_title_key,
				self._ticket_id,
				self._ec_pubkey,
				device_cert
			)
			self._personal_enc_title_key = self._enc_title_key
			self._enc_title_key = key
			self._tik_block = self._tik_block[:0x7F] + key + self._tik_block[0x8F:]
		else:
			self._personal_enc_title_key = self._encrypt_personal_tik_key_layer(
				self._enc_title_key,
				self._ticket_id,
				self._ec_pubkey,
				device_cert
			)

		return True

	def has_rights_to_index(self, index: typing.SupportsInt) -> bool:
		index = int(index)
		if index >= 65536:
			return False
		return (self._rights & (1 << index)) != 0



from enum import IntEnum, unique
from struct import pack, unpack
import typing

from cryptography.hazmat.primitives.asymmetric import utils as crypto_utils, padding as crypto_padding
from cryptography.hazmat.primitives.asymmetric import ec, rsa
from cryptography.hazmat.primitives import hashes
from cryptography.exceptions import InvalidSignature as crypto_InvalidSignature

from .exception import CTRExceptionBase, ClassInitError

__all__ = [
	"DigitalSignature",
	"SignatureType",
	"KeyType",
	"InvalidSignatureError",
	"SignatureProcessingError",
	"SignatureTypeError",
	"VerificationError",
	"IssuingError",
	"KeyTypeError"
]

_PubkeysType = typing.Union[
	ec.EllipticCurvePublicKey,
	ec.EllipticCurvePublicKeyWithSerialization,
	rsa.RSAPublicKey,
	rsa.RSAPublicKeyWithSerialization
]

_PrivkeysType = typing.Union[
	ec.EllipticCurvePrivateKey,
	ec.EllipticCurvePrivateKeyWithSerialization,
	rsa.RSAPrivateKey,
	rsa.RSAPrivateKeyWithSerialization
]

class InvalidSignatureError(CTRExceptionBase):
	"""General DigitalSignature invalid signature exception"""

class SignatureProcessingError(CTRExceptionBase):
	"""General DigitalSignature signature check exception"""

class SignatureTypeError(CTRExceptionBase):
	"""General SignatureType exception"""

class VerificationError(CTRExceptionBase):
	"""General IssuedObject verification exception"""

class IssuingError(CTRExceptionBase):
	"""General IssuedObject reissuing exception"""

class KeyTypeError(CTRExceptionBase):
	"""General KeyType exception"""

@unique
class SignatureType(IntEnum):
	RSA_4096_SHA1   = 0x10000
	RSA_2048_SHA1   = 0x10001
	ECC_SHA1        = 0x10002
	RSA_4096_SHA256 = 0x10003
	RSA_2048_SHA256 = 0x10004
	ECC_SHA256      = 0x10005

	@classmethod
	def is_valid(cls, entry: typing.Union['SignatureType', int]) -> bool:
		value = int(entry)
		return value >= cls.RSA_4096_SHA1 and value <= cls.ECC_SHA256

	@classmethod
	def sign_block_size(cls, entry: typing.Union['SignatureType', int]) -> int:
		value = int(entry)
		if value == cls.RSA_4096_SHA1 or value == cls.RSA_4096_SHA256:
			return 0x200
		if value == cls.RSA_2048_SHA1 or value == cls.RSA_2048_SHA256:
			return 0x100
		if value == cls.ECC_SHA1 or value == cls.ECC_SHA256:
			return 0x3C
		raise SignatureTypeError("Invalid SignatureType")

	@classmethod
	def sign_padding_block_size(cls, entry: typing.Union['SignatureType', int]) -> int:
		value = int(entry)
		if value == cls.RSA_4096_SHA1 or value == cls.RSA_4096_SHA256 or value == cls.RSA_2048_SHA1 or value == cls.RSA_2048_SHA256:
			return 0x3C
		if value == cls.ECC_SHA1 or value == cls.ECC_SHA256:
			return 0x40
		raise SignatureTypeError("Invalid SignatureType")

	@classmethod
	def hash_algo(cls, entry: typing.Union['SignatureType', int]) -> hashes.HashAlgorithm:
		value = int(entry)
		if value == cls.RSA_4096_SHA1 or value == cls.RSA_2048_SHA1 or value == cls.ECC_SHA1:
			return hashes.SHA1
		if value == cls.RSA_4096_SHA256 or value == cls.RSA_2048_SHA256 or value == cls.ECC_SHA256:
			return hashes.SHA256
		raise SignatureTypeError("Invalid SignatureType")

@unique
class KeyType(IntEnum):
	RSA_4096 = 0
	RSA_2048 = 1
	ECC      = 2

	@classmethod
	def is_valid(cls, entry: typing.Union['KeyType', int]) -> bool:
		value = int(entry)
		return value >= cls.RSA_4096 and value <= cls.ECC

	@classmethod
	def does_verify_digital_sign_type(
		cls,
		certtype: typing.Union['KeyType', int],
		digsigntype: typing.Union[SignatureType, int]
	) -> bool:
		certtype = cls(certtype)
		digsigntype = SignatureType(digsigntype)
		if (digsigntype == SignatureType.RSA_4096_SHA1 or digsigntype == SignatureType.RSA_4096_SHA256) and certtype == cls.RSA_4096:
			return True
		if (digsigntype == SignatureType.RSA_2048_SHA1 or digsigntype == SignatureType.RSA_2048_SHA256) and certtype == cls.RSA_2048:
			return True
		if (digsigntype == SignatureType.ECC_SHA1 or digsigntype == SignatureType.ECC_SHA256) and certtype == cls.ECC:
			return True
		return False

class DigitalSignature:
	def __init__(self, data: typing.SupportsBytes):
		self._sign_type = unpack(">I", bytes(data)[:4])[0]

		if not SignatureType.is_valid(self._sign_type):
			raise ClassInitError("Invalid signature type")

		block_size = SignatureType.sign_block_size(self._sign_type)

		self._sign_block = bytes(data[4:block_size+4])

		if len(self._sign_block) != block_size:
			raise ClassInitError("Not enough data to get the full block size")

	def __bytes__(self):
		block_size = SignatureType.sign_block_size(self._sign_type)
		pad_size = SignatureType.sign_padding_block_size(self._sign_type)

		return pack(f">I{block_size}s{pad_size}x", self._sign_type, self._sign_block)

	def __len__(self):
		block_size = SignatureType.sign_block_size(self._sign_type)
		pad_size = SignatureType.sign_padding_block_size(self._sign_type)
		return 4 + block_size + pad_size

	def __repr__(self):
		modname = self.__class__.__module__
		clsname = self.__class__.__qualname__
		sigtype = SignatureType(self._sign_type)
		return f"<{modname}.{clsname} sign_type={repr(sigtype)}>"

	@property
	def sign_type(self) -> SignatureType:
		return SignatureType(self._sign_type)

	@property
	def raw_sign_block(self) -> bytes:
		return self._sign_block

	@property
	def encoded_signature(self) -> bytes:
		block_size = SignatureType.sign_block_size(self._sign_type)

		if block_size == 0x3C:
			r, s = unpack("30s30s", self._sign_block)
			r = int(r.hex(), 16)
			s = int(s.hex(), 16)
			return crypto_utils.encode_dss_signature(r, s)

		return self.raw_sign_block

	@property
	def hash_algo(self) -> hashes.HashAlgorithm:
		return SignatureType.hash_algo(self._sign_type)

	def _pubkey_validate(
		self,
		key: _PubkeysType
	) -> typing.Optional[_PubkeysType]:
		block_size = SignatureType.sign_block_size(self._sign_type)

		if block_size == 0x3C:
			if isinstance(key, ec.EllipticCurvePrivateKeyWithSerialization) or isinstance(key, ec.EllipticCurvePrivateKey):
				return key.public_key()
			if isinstance(key, ec.EllipticCurvePublicKeyWithSerialization) or isinstance(key, ec.EllipticCurvePublicKey):
				return key
			return None

		if isinstance(key, rsa.RSAPrivateKeyWithSerialization) or isinstance(key, rsa.RSAPrivateKey):
			return key.public_key()
		if isinstance(key, rsa.RSAPublicKeyWithSerialization) or isinstance(key, rsa.RSAPublicKey):
			return key
		return None

	def _privkey_validate(
		self,
		key: _PrivkeysType,
		signtype: typing.Union[SignatureType, int]
	) -> typing.Optional[_PrivkeysType]:
		block_size = SignatureType.sign_block_size(signtype)

		if block_size == 0x3C:
			if isinstance(key, ec.EllipticCurvePrivateKeyWithSerialization) or isinstance(key, ec.EllipticCurvePrivateKey):
				return key
			return None

		if isinstance(key, rsa.RSAPrivateKeyWithSerialization) or isinstance(key, rsa.RSAPrivateKey):
			return key
		return None

	def verify_data(
		self,
		key: _PubkeysType,
		data: typing.SupportsBytes
	) -> None:
		try:
			hashalgo = self.hash_algo()

			key = self._pubkey_validate(key)

			if SignatureType.sign_block_size(self._sign_type) == 0x3C:
				key.verify(self.encoded_signature, data, ec.ECDSA(hashalgo))
			else:
				key.verify(self.encoded_signature, data, crypto_padding.PKCS1v15(), hashalgo)

		except crypto_InvalidSignature as e:
			raise InvalidSignatureError("Invalid Signature") from e
		except (TypeError, AttributeError) as e:
			raise SignatureProcessingError("Unexpected exception or invalid key object") from e
		except Exception as e:
			raise SignatureProcessingError("Unexpected exception") from e

	def verify_digest(
		self,
		key: _PubkeysType,
		digest: typing.SupportsBytes
	) -> None:
		try:
			hashalgo = self.hash_algo()

			key = self._pubkey_validate(key)

			if SignatureType.sign_block_size(self._sign_type) == 0x3C:
				key.verify(self.encoded_signature, digest, ec.ECDSA(crypto_utils.Prehashed(hashalgo)))
			else:
				key.verify(self.encoded_signature, digest, crypto_padding.PKCS1v15(), crypto_utils.Prehashed(hashalgo))

		except crypto_InvalidSignature as e:
			raise InvalidSignatureError("Invalid Signature") from e
		except (TypeError, AttributeError) as e:
			raise SignatureProcessingError("Unexpected exception or invalid key object") from e
		except Exception as e:
			raise SignatureProcessingError("Unexpected exception") from e

	def update_sign_with_data(
		self,
		signtype: typing.Union[SignatureType, int, None],
		key: _PrivkeysType,
		data: typing.SupportsBytes
	) -> None:
		try:
			if not signtype:
				signtype = self._sign_type

			hashalgo = SignatureType.hash_algo(signtype)

			key = self._privkey_validate(key, signtype)

			block_size = SignatureType.sign_block_size(signtype)

			if block_size == 0x3C:
				signature = key.sign(bytes(data), ec.ECDSA(hashalgo()))
				signature = crypto_utils.decode_dss_signature(signature)
				signature = pack("30s30s", signature[0].to_bytes(30, 'big'), signature[1].to_bytes(30, 'big'))
			else:
				signature = key.sign(bytes(data), crypto_padding.PKCS1v15(), hashalgo())

			if len(signature) != block_size:
				raise SignatureProcessingError("Key did not generate signature of proper length for requested signature type")

			self._sign_block = signature

		except (TypeError, AttributeError) as e:
			raise SignatureProcessingError("Unexpected exception or invalid key object") from e
		except SignatureProcessingError:
			raise
		except Exception as e:
			raise SignatureProcessingError("Unexpected exception") from e

	def update_sign_with_digest(
		self,
		signtype: typing.Union[SignatureType, int, None],
		key: _PrivkeysType,
		digest: typing.SupportsBytes
	) -> None:
		try:
			if not signtype:
				signtype = self._sign_type

			hashalgo = SignatureType.hash_algo(signtype)

			key = self._privkey_validate(key, signtype)

			block_size = SignatureType.sign_block_size(signtype)

			if block_size == 0x3C:
				signature = key.sign(bytes(digest), ec.ECDSA(crypto_utils.Prehashed(hashalgo())))
				signature = crypto_utils.decode_dss_signature(signature)
				signature = pack("30s30s", signature[0].to_bytes(30, 'big'), signature[1].to_bytes(30, 'big'))
			else:
				signature = key.sign(bytes(digest), crypto_padding.PKCS1v15(), crypto_utils.Prehashed(hashalgo()))

			if len(signature) != block_size:
				raise SignatureProcessingError("Key did not generate signature of proper length for requested signature type")

			self._sign_block = signature

		except (TypeError, AttributeError) as e:
			raise SignatureProcessingError("Unexpected exception or invalid key object") from e
		except SignatureProcessingError:
			raise
		except Exception as e:
			raise SignatureProcessingError("Unexpected exception") from e

class IssuingObject:
	def __init__(
		self,
		callable_get_key_type: typing.Callable[[], KeyType],
		callable_get_pub_key: typing.Callable[
			[],
			_PubkeysType
		],
		callable_get_priv_key: typing.Callable[
			[],
			typing.Optional[_PrivkeysType]
		],
		callable_get_issuing_name: typing.Callable[[], str]
	):
		self._issuer_cbs = {
			'key_type': callable_get_key_type,
			'pub_key': callable_get_pub_key,
			'priv_key': callable_get_priv_key,
			'issuing_name': callable_get_issuing_name
		}

class IssuedObject:
	def __init__(
		self,
		callable_get_data: typing.Callable[[], typing.SupportsBytes],
		callable_get_digital_sign: typing.Callable[[], DigitalSignature],
		callable_get_issuer: typing.Callable[[], str],
		callable_get_reissued_data: typing.Callable[[str], typing.SupportsBytes],
		callable_update: typing.Callable[[DigitalSignature, str, typing.SupportsBytes], None]
	):
		self._issued_cbs = {
			'data': callable_get_data,
			'sign': callable_get_digital_sign,
			'issuer': callable_get_issuer,
			'data_for_resign_reissue': callable_get_reissued_data,
			'update_reissued': callable_update
		}

	def verify(
		self,
		issuer_or_key: typing.Union[
			IssuingObject,
			_PubkeysType
		]
	) -> None:
		_sign = self._issued_cbs['sign']()
		if isinstance(issuer_or_key, IssuingObject):
			issuer = issuer_or_key

			if issuer._issuer_cbs['issuing_name']() != self._issued_cbs['issuer']():
				raise VerificationError("Given issuing object does not verify this issued object")

			if not KeyType.does_verify_digital_sign_type(issuer._issuer_cbs['key_type'](), _sign.sign_type):
				raise VerificationError("Given issuing object key type does not verify this issued object")

			key = issuer._issuer_cbs['pub_key']()
		else:
			key = issuer_or_key

		_sign.verify_data(key, self._issued_cbs['data']())

	def reissue(
		self,
		issuer_or_key: typing.Union[
			IssuingObject,
			_PrivkeysType
		],
		signtype: typing.Union[SignatureType, int, None] = None,
		issuer: typing.Optional[str] = None
	) -> None:
		if isinstance(issuer_or_key, IssuingObject):
			issuer = issuer_or_key

			issuer_str = issuer._issuer_cbs['issuing_name']()

			key = issuer._issuer_cbs['priv_key']()
		else:
			key = issuer_or_key

			try:
				issuer_str = str(issuer_str) if issuer_str is not None else None
			except Exception as e:
				raise IssuingError("Issuer not string") from e

		if not key:
			raise IssuingError("Private key not available")

		if not issuer_str:
			raise IssuingError("No issuer name")

		if len(issuer_str) > 64:
			raise IssuingError("Issuer name too big, max 64")

		_new_data = self._issued_cbs['data_for_resign_reissue'](issuer_str)

		_sign_copy = DigitalSignature(bytes(self._issued_cbs['sign']()))
		_sign_copy.update_sign_with_data(signtype, key, _new_data)

		self._issued_cbs['update_reissued'](_sign_copy, issuer_str, _new_data)



import pycurl, typing, re
from collections import OrderedDict

from .exception import CleanintyExceptionBase, ClassInitError

__all__ = [
	"Connection",
	"SimpleDownloadBuffer",
	"MultiWriter",
	"WriterFuncType",
	"ParameterSetError"
]

WriterFuncType = typing.Callable[
	[typing.SupportsBytes], None
]

class ParameterSetError(CleanintyExceptionBase):
	"""General Connection parameter setting exception"""

class SimpleDownloadBuffer:
	def __init__(self):
		self._data = b''
		self._data_chunks = []

	def write(self, x: typing.SupportsBytes) -> None:
		self._data_chunks.append(bytes(x))

	def get(self) -> bytes:
		if self._data_chunks:
			self._data += b''.join(self._data_chunks)
			self._data_chunks.clear()
		return self._data

class MultiWriter:
	def __init__(self, writers: typing.Iterable[WriterFuncType]):
		self._writers = []
		for i in writers:
			if not callable(i):
				raise ClassInitError("non callable on a callable list!")
			self._writers.append(i)
		self._writers = tuple(self._writers)

	def write(self, x: typing.SupportsBytes) -> None:
		for i in self._writers:
			i(x)

class Connection:
	def __init__(self, *, max_connects: int = 1):
		self._headers = []
		self._keepalive = False
		try:
			self._curl_handle = pycurl.Curl()
			self._curl_handle.setopt(pycurl.HTTP_VERSION, pycurl.CURL_HTTP_VERSION_1_1)
			self._curl_handle.setopt(pycurl.MAXCONNECTS, max_connects)
			self._curl_handle.setopt(pycurl.TCP_KEEPALIVE, 0)
		except Exception as e:
			raise ClassInitError("Failed to initialize curl handle") from e

	def __del__(self):
		self.close()

	def _validate_header(self, name, value):
		if not isinstance(name, str) or \
			not re.fullmatch("^[A-Za-z0-9!#$%&'*+-.^_`|~]+$", name):
			return False
		if value is not None and \
			(not isinstance(value, str) or \
				(value and not re.fullmatch("^[\\x09\\x20-\\x7e]+$", value))
			):
			return False
		return True

	def _get_header_list(self):
		headers = []
		default_overrides = OrderedDict()
		default_overrides['user-agent'] = 'User-Agent:'
		default_overrides['accept'] = 'Accept:'

		for v in self._headers:
			if isinstance(v[1], str) and len(v[1]) != 0:
				header = v[0] + ': ' + v[1]
			elif v[1] is None:
				header = v[0] + ':'
			else:
				header = v[0] + ';'
			headers.append(header)
			default_overrides.pop(v[0].lower(), None)

		headers = list(default_overrides.values()) + headers

		return headers

	def set_keepalive(
		self,
		keepalive: bool,
		keepidle: typing.SupportsInt = 120,
		keepintvl: typing.SupportsInt = 60
	) -> None:
		try:
			if keepalive:
				keepidle = int(keepidle)
				keepintvl = int(keepintvl)

				if keepidle < 0 and keepintvl < 0:
					raise ParameterSetError("keepidle or keepintvl can't be negative")

				self._curl_handle.setopt(pycurl.TCP_KEEPIDLE, keepidle)
				self._curl_handle.setopt(pycurl.TCP_KEEPINTVL, keepintvl)
				self._curl_handle.setopt(pycurl.TCP_KEEPALIVE, 1)
				self._keepalive = True

			else:
				self._curl_handle.setopt(pycurl.TCP_KEEPALIVE, 0)
				self._keepalive = False

		except Exception as e:
			raise ParameterSetError("Failed to set keep-alive option") from e

	def set_header(self, name: str, value: typing.Optional[str]) -> None:
		if not self._validate_header(name, value):
			raise ParameterSetError("Could not validate field name or value")

		if value is None:
			self.remove_header(name, all_elements = True)
			return

		self._headers.append([name, value])

	def reset_headers(self):
		self._headers.clear()

	def remove_header(
		self,
		name: str,
		*,
		latest_element: bool = True,
		all_elements: bool = False
	) -> typing.Optional[typing.List[str]]:
		if not self._validate_header(name, None):
			raise ParameterSetError("Could not validate field name")

		_name = name.lower()

		_headers = self._headers.copy()
		if latest_element or all_elements:
			_headers.reverse()

		if all_elements:
			for i, j in enumerate(_headers):
				if j[0].lower() == _name:
					self._headers.pop((len(_headers) - i - 1))
			return None

		index = -1
		for i, j in enumerate(_headers):
			if j[0].lower() != _name:
				continue

			index = (len(_headers) - i - 1) if latest_element else i
			break

		return self._headers.pop(index) if index >= 0 else None

	def set_cainfo(self, path: typing.Optional[str]) -> None:
		try:
			self._curl_handle.setopt(pycurl.CAINFO, path)
		except Exception as e:
			raise ParameterSetError("Failed setting CAINFO") from e

	def set_cli_cert(self, cert_path: typing.Optional[str], key_path: typing.Optional[str]) -> None:
		try:
			self._curl_handle.setopt(pycurl.SSLCERT, cert_path)
			self._curl_handle.setopt(pycurl.SSLKEY, key_path)
		except Exception as e:
			raise ParameterSetError("Failed setting client certificate") from e

	def set_url(self, url: str) -> None:
		try:
			self._curl_handle.setopt(pycurl.URL, str(url))
		except Exception as e:
			raise ParameterSetError("Failed setting url") from e

	def set_customrequest(self, request: typing.Optional[str] = None):
		try:
			self._curl_handle.setopt(pycurl.CUSTOMREQUEST, str(request))
		except Exception as e:
			raise ParameterSetError("Failed setting custom request") from e

	def set_verbose(self, toggle: bool) -> None:
		try:
			self._curl_handle.setopt(pycurl.VERBOSE, 1 if toggle else 0)
		except Exception as e:
			raise ParameterSetError("Failed setting verbose") from e

	def set_post_data(self, data: typing.Union[str, bytes, None]) -> None:
		try:
			if data:
				self._curl_handle.setopt(pycurl.POSTFIELDSIZE, len(data))
				self._curl_handle.setopt(pycurl.POSTFIELDS, data)
				self._data = data
			else:
				self._curl_handle.setopt(pycurl.POST, 0)
				self._data = None
		except Exception as e:
			raise ParameterSetError("Failed setting post data") from e

	def set_write_function(self, func: typing.Optional[WriterFuncType]) -> None:
		try:
			self._curl_handle.setopt(pycurl.WRITEFUNCTION, func)
		except Exception as e:
			raise ParameterSetError("Failed setting write function") from e

	def set_write_header_function(self, func: typing.Optional[WriterFuncType]) -> None:
		try:
			self._curl_handle.setopt(pycurl.HEADERFUNCTION, func)
		except Exception as e:
			raise ParameterSetError("Failed setting write function") from e

	def set_nobody(self, toggle: bool) -> None:
		try:
			self._curl_handle.setopt(pycurl.NOBODY, 1 if toggle else 0)
		except Exception as e:
			raise ParameterSetError("Failed setting nobody") from e

	def perform(self) -> int:
		headers = self._get_header_list()
		try:
			self._curl_handle.setopt(pycurl.HTTPHEADER, headers)
		except Exception as e:
			raise ParameterSetError("Failed setting headers") from e

		self._curl_handle.perform()

		return self._curl_handle.getinfo(pycurl.RESPONSE_CODE)

	def close(self) -> None:
		self._curl_handle.close()

	@property
	def is_keepalive(self) -> bool:
		return self._keepalive



import base64, typing

from ...nintendowifi import soapenvelopebase
from .manager import CtrSoapManager
from ..exception import ClassInitError, DataProcessingError
from .exception import OperationError, CTRExceptionBase
from ..title import Title
from ._common_parsers import _parse_attribute_member

#TODO:
# - ListContentSetGroups
# ugh, I really dont like the next ones
# - ListECardItems
# - ListTitlesEx
# - ListContentSetsEx
# - ListItems

__all__ = [
	"GetContentSizes",
	"GetCountryAttributes",
	"GetCountryMigrateAttributes",
	"ListCashReplenishAmounts"
]

class CatalogContentSizes:
	def __init__(
		self,
		content_id: int,
		content_index: int,
		size: int
	):
		if not 0 <= content_id <= 0xffffffff or \
			not 0 <= content_index < 65536 or \
			not 0 <= size <= 0xffffffffffffffff:
			raise DataProcessingError("Invalid content size information!")

		self._id = content_id
		self._index = content_index
		self._size = size

	@property
	def id(self) -> int:
		return self._id

	@property
	def index(self) -> int:
		return self._index

	@property
	def size(self) -> int:
		return self._size

def _xml_get_contentsize(
	parent: soapenvelopebase.SoapEnvelopeBase,
	element: soapenvelopebase.XML_Element
) -> CatalogContentSizes:
	text = parent._xml_get_str_element(parent, element)
	split = text.split(',', 3) # i expect 2 commas, but 3 lets me see if too many
	if len(split) != 3:
		raise DataProcessingError("Not enough or too many commas on ContentSize!")
	return CatalogContentSizes(int(split[0], 16), int(split[1]), int(split[2]))

# not for direct use with _xml_xxx_parse
def _xml_get_str_len_limited(
	parent: soapenvelopebase.SoapEnvelopeBase,
	element: soapenvelopebase.XML_Element,
	length: int
) -> str:
	text = parent._xml_get_str_element(parent, element)
	if len(text) > length:
		raise DataProcessingError("str element surpassed length limits")
	return text

class GetContentSizes(soapenvelopebase.SoapEnvelopeBase):
	@soapenvelopebase.ObjectTimingEmuHelper(0.9, 0.125)
	def __init__(
		self,
		ctrsoapmanager: CtrSoapManager,
		title: Title,
		content_indexes: typing.Iterable[typing.SupportsInt],
		resort: bool = False
	):
		if not isinstance(ctrsoapmanager, CtrSoapManager):
			raise ClassInitError("Expected CtrSoapManager")

		if not isinstance(title, Title):
			raise ClassInitError("Expected Title for title")

		# there's also a pre-requisite to use ctrsoapmanager.set_app_and_tin to set application id and tin, its a must!

		super().__init__(soapenvelopebase.SoapSubNames.CAS, 'GetContentSizes', ctrsoapmanager, False, False)

		self._write_tag('TitleId', f"{title.id:016X}")
		self._write_tag('TitleVersion', f"{title.version}")

		count = 0
		indexes = []
		for i in content_indexes:
			i = int(i)
			if not 0 <= i < 65536:
				raise DataProcessingError("Content indexes are an u16 value")
			indexes.append(f"{i}")
			if len(indexes) == 100:
				self._write_tag('ContentIndex', ','.join(indexes))
				indexes.clear()
			count += 1
		if indexes:
			self._write_tag('ContentIndex', ','.join(indexes))

		del indexes

		ret = self._send(ctrsoapmanager.get_url_by_identifier('cas'))
		if ret != 200:
			raise OperationError("Bad HTTP response or connection error, ret = {0}".format(ret))

		response_parse = self._initiate_response_parse()
		self._validate_errorcode(self.errorcode, self.errormessage)
		try:
			response = response_parse[1]
			self._contentsizes = self._xml_multi_element_parse(response, 'urn:ContentSize', _xml_get_contentsize)
			self._contentsizes = tuple(self._contentsizes) if self._contentsizes is not None else tuple()
			self._tmdsize = self._xml_element_parse(response, 'urn:TMDSize', self._xml_get_int_element)
			if not 0 <= self._tmdsize < 0xffffffff:
				raise DataProcessingError("Invalid TMD Size!")
			self._maxcontentindex = self._xml_element_parse(response, 'urn:MaxContentIndex', self._xml_get_int_element)
			if not 0 <= self._maxcontentindex < 65536:
				raise DataProcessingError("Out of range max content index!")
			# it's expected to return same amount of content sizes, no more no less, but I won't error here
			self._got_all = count == len(self._contentsizes)
		except CTRExceptionBase:
			raise
		except Exception as e:
			raise soapenvelopebase.XMLParseError("Unexpected exception while parsing XML") from e

		if resort:
			self._contentsizes = tuple(sorted(self._contentsizes, key=lambda x: x.index))

	@property
	def contentsizes(self) -> typing.Iterable[CatalogContentSizes]:
		return self._contentsizes

	@property
	def tmdsize(self) -> int:
		return self._tmdsize
	
	@property
	def maxcontentindex(self) -> int:
		return self._maxcontentindex

	@property
	def got_all(self) -> bool:
		return self._got_all # may be False if indexes list was empty, returns all of them if none specified on Soap

class GetCountryAttributes(soapenvelopebase.SoapEnvelopeBase):
	@soapenvelopebase.ObjectTimingEmuHelper(0.9, 0.125)
	def __init__(self, ctrsoapmanager: CtrSoapManager):
		if not isinstance(ctrsoapmanager, CtrSoapManager):
			raise ClassInitError("Expected CtrSoapManager")

		super().__init__(soapenvelopebase.SoapSubNames.CAS, 'GetCountryAttributes', ctrsoapmanager, False, False)

		ret = self._send(ctrsoapmanager.get_url_by_identifier('cas'))
		if ret != 200:
			raise OperationError("Bad HTTP response or connection error, ret = {0}".format(ret))

		response_parse = self._initiate_response_parse()
		self._validate_errorcode(self.errorcode, self.errormessage)
		try:
			response = response_parse[1]
			self._countryattribits = self._xml_element_parse(response, 'urn:CountryAttribits', self._xml_get_int_element, True)
			self._countryattribits = 0 if self._countryattribits is None else self._countryattribits
			if self._countryattribits < 0 or self._countryattribits > 0xffffffffffffffff:
				raise DataProcessingError("Out of range country attribits")
			self._countryattributes = self._xml_multi_element_parse(response, 'urn:CountryAttributes', _parse_attribute_member, True)
			self._countryattributes = tuple(self._countryattributes) if self._countryattributes is not None else tuple()
		except CTRExceptionBase:
			raise
		except Exception as e:
			raise soapenvelopebase.XMLParseError("Unexpected exception while parsing XML") from e

	@property
	def countryattribits(self) -> int:
		return self._countryattribits

	@property
	def countryattributes(self) -> typing.Iterator[typing.Tuple[str, str]]:
		return self._countryattributes

class GetCountryMigrateAttributes(soapenvelopebase.SoapEnvelopeBase):
	@soapenvelopebase.ObjectTimingEmuHelper(0.9, 0.125)
	def __init__(self, ctrsoapmanager: CtrSoapManager):
		if not isinstance(ctrsoapmanager, CtrSoapManager):
			raise ClassInitError("Expected CtrSoapManager")

		super().__init__(soapenvelopebase.SoapSubNames.CAS, 'GetCountryMigrateAttributes', ctrsoapmanager, False, False)

		ret = self._send(ctrsoapmanager.get_url_by_identifier('cas'))
		if ret != 200:
			raise OperationError("Bad HTTP response or connection error, ret = {0}".format(ret))

		response_parse = self._initiate_response_parse()
		self._validate_errorcode(self.errorcode, self.errormessage)
		try:
			response = response_parse[1]
			self._enablemigratetitle = self._xml_element_parse(response, 'urn:EnableMigrateTitle', self._xml_get_bool_element)
			self._enablemigrateaccount = self._xml_element_parse(response, 'urn:EnableMigrateAccount', self._xml_get_bool_element)
		except CTRExceptionBase:
			raise
		except Exception as e:
			raise soapenvelopebase.XMLParseError("Unexpected exception while parsing XML") from e

	@property
	def enablemigratetitle(self) -> bool:
		return self._enablemigratetitle

	@property
	def enablemigrateaccount(self) -> bool:
		return self._enablemigrateaccount

class ListCashReplenishAmounts(soapenvelopebase.SoapEnvelopeBase):
	@soapenvelopebase.ObjectTimingEmuHelper(0.9, 0.125)
	def __init__(self, ctrsoapmanager: CtrSoapManager):
		if not isinstance(ctrsoapmanager, CtrSoapManager):
			raise ClassInitError("Expected CtrSoapManager")

		super().__init__(soapenvelopebase.SoapSubNames.CAS, 'ListCashReplenishAmounts', ctrsoapmanager, False, True)

		ret = self._send(ctrsoapmanager.get_url_by_identifier('cas'))
		if ret != 200:
			raise OperationError("Bad HTTP response or connection error, ret = {0}".format(ret))

		response_parse = self._initiate_response_parse()
		self._validate_errorcode(self.errorcode, self.errormessage)
		try:
			response = response_parse[1]
			get_capped_str = lambda x,y: _xml_get_str_len_limited(x, y, 31)
			self._currency = self._xml_element_parse(response, 'urn:Currency', get_capped_str)
			self._mincash = self._xml_element_parse(response, 'urn:MinCash', get_capped_str)
			self._maxcashbalance = self._xml_element_parse(response, 'urn:MaxCashBalance', get_capped_str)
			self._listamounts = self._xml_multi_element_parse(response, 'urn:ListAmounts', get_capped_str, True)
			self._listamounts = tuple(self._listamounts) if self._listamounts is not None else tuple()
		except CTRExceptionBase:
			raise
		except Exception as e:
			raise soapenvelopebase.XMLParseError("Unexpected exception while parsing XML") from e

	@property
	def currency(self) -> str:
		return self._currency

	@property
	def mincash(self) -> str:
		return self._mincash

	@property
	def maxcashbalance(self) -> str:
		return self._maxcashbalance

	@property
	def listamounts(self) -> typing.Iterable[str]:
		return self._listamounts
